#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation phase requirement sets** (i.e., immutable sets of simulation phase
requirements, requiring zero or more arbitrary requirements to be satisfied).

Design
----------
This submodule *only* collects submodule-scope constants, all of which are
instances of the :class:`SimPhaseRequirements` class.
'''

#FIXME: Generalize the requirement globals defined below to be more globally
#usable. Currently, they require the current "phase" object be passed to their
#is_satisfied() methods, which is cumbersome at best. To rectify this, first
#note that whether or not a given phase satisfies a given requirement is a
#constant that should *NEVER* change during that phase. Hence, we can optimally
#associate a given phase with all of the following requirements as follows:
#
#* Define a new "SimPhaseRequirements" class in this submodule.
#* Define a new "SimPhase.reqs" instance variable whose value is an instance of
#  this "SimPhaseRequirements" class.
#* Define a SimPhaseRequirements.__init__() method:
#  * Accepting only the current "SimPhase" object.
#  * For each requirement global defined below (e.g., "DEFORM"):
#    * Defining a boolean instance variable of the same name, lowercased and
#      prefixed by "is_", whose value is the value of the corresponding
#      requirement global passed this phase.
#
#For example, as a first-draft implementation:
#
# ....................{ CLASSES                           }....................
# #FIXME: Document us up.
# class SimPhaseRequirements(object):
#
#     def __init__(self, phase: SimPhase) -> None:
#
#         self.is_deform = DEFORM(phase)
#
#Pretty sleek, eh? While these instance variables could also be defined as
#read-only properties, doing so is significantly more cumbersome and slightly
#less efficient. So, why bother? There are too many requirements to make doing
#so worthwhile or sane. (Yay!)

# ....................{ IMPORTS                           }....................
from betse.science.phase.require import phasereqsmake
from betse.science.phase.require.abc.phasereqset import SimPhaseRequirements
from betse.science.enum.enumconf import SolverType

# ....................{ REQUIREMENTS                      }....................
NONE = SimPhaseRequirements()
'''
**Null requirements** (i.e., requirements unconditionally satisfied by all
simulation phases regardless of which simulation features those phases enable).
'''


ELECTRIC_CURRENT = NONE
'''
Requirements that a simulation phase enable electrical currents and hence at
least intracellular electrical currents.

Simulation phases satisfying this requirement do *not* necessarily enable
extracellular electrical currents, which additionally require the full solver
to be enabled.
'''


ELECTRIC_FIELD = NONE
'''
Requirements that a simulation phase enable electrical fields and hence at
least intracellular electrical fields.

Simulation phases satisfying this requirement do *not* necessarily enable
extracellular electrical fields, which additionally require the full solver to
be enabled.
'''


MICROTUBULE = NONE
'''
Requirements that a simulation phase enable microtubules.
'''

# ....................{ REQUIREMENTS ~ bool               }....................
# Requirements satisfied by enabling a requirement-specific boolean.

ECM = phasereqsmake.make_requirements_bool_expr(
    name='extracellular spaces', bool_expr='phase.p.is_ecm')
'''
Requirements that a simulation phase enable the extracellular matrix (ECM),
also referred to as "extracellular spaces."

Note that this requirement does *not* itself require the full BETSE solver.
While the equivalent circuit-based solver mostly ignores extracellular spaces,
gene regulatory networks (GRNs) supported by this solver explicitly support
extracellular spaces. Ergo, this requirement is solver-agnostic.
'''

# ....................{ REQUIREMENTS ~ solver             }....................
# This solver is internally referenced by solver classes instantiated below
# (e.g., "SimPhaseRequirementsIon", "SimPhaseRequirementsSolverFullAnd") and
# *MUST* thus be declared first -- before these classes are instantiated.

SOLVER_FULL = phasereqsmake.make_requirements_enum_expr(
    name='full BETSE solver',
    enum_expr='phase.p.solver_type',
    enum_member=SolverType.FULL,
)
'''
Requirements that a simulation phase enable the full BETSE solver.
'''

# ....................{ REQUIREMENTS ~ solver : equal     }....................
# Requirements satisfied by enabling only the full solver. Technically, these
# requirements are currently identical to the previously defined "SOLVER_FULL"
# requirement. This need *NOT* necessarily be the case, however. Theoretically,
# subsequent versions of this application could expose boolean settings in the
# simulation configuration file permitting these currently unconditional
# simulation features to be selectively disabled.


ELECTRIC_CURRENT_EXTRA = SOLVER_FULL
'''
Requirements that a simulation phase enable extracellular electrical currents.

Simulation phases satisfying this requirement necessarily enable the full
solver but *not* extracellular spaces, which are *not* required to simulate
extracellular electrical currents.
'''


ELECTRIC_CURRENT_MEMBRANE = SOLVER_FULL
'''
Requirements that a simulation phase enable transmembrane electrical currents.

Simulation phases satisfying this requirement necessarily enable the full
solver but *not* extracellular spaces, which are *not* required to simulate
transmembrane electrical currents.
'''


PUMP_NAKATPASE = SOLVER_FULL
'''
Requirements that a simulation phase enable the Na-K-ATPase membrane pump.
'''

# ....................{ REQUIREMENTS ~ solver : equal     }....................
# Requirements satisfied by enabling only both the full solver *AND*
# extracellular spaces. See above for comments.

SOLVER_FULL_ECM = SOLVER_FULL | ECM
'''
Requirements that a simulation phase enable both extracellular spaces *and* a
solver able to fully simulate these spaces (e.g., the full BETSE solver).
'''


ELECTRIC_FIELD_EXTRA = SOLVER_FULL_ECM
'''
Requirements that a simulation phase enable extracellular electrical fields and
hence both extracellular spaces *and* a solver capable of fully simulating
these spaces (e.g., the full BETSE solver).
'''


VOLTAGE_EXTRA = SOLVER_FULL
'''
Requirements that a simulation phase enable extracellular voltages and hence
both extracellular spaces *and* a solver capable of fully simulating these
spaces (e.g., the full BETSE solver).
'''

# ....................{ REQUIREMENTS ~ solver : bool      }....................
# Requirements satisfied by enabling only the full solver and a
# requirement-specific boolean.

DEFORM = SOLVER_FULL | phasereqsmake.make_requirements_bool_expr(
    name='cellular deformation', bool_expr='phase.p.deformation')
'''
Requirements that a simulation phase enable cellular deformations.
'''


ELECTROOSMOSIS = SOLVER_FULL | phasereqsmake.make_requirements_bool_expr(
    name='electroosmotic flow', bool_expr='phase.p.sim_eosmosis')
'''
Requirements that a simulation phase enable electroosmotic flow (EOF).
'''


FLUID = SOLVER_FULL | phasereqsmake.make_requirements_bool_expr(
    name='fluid flow', bool_expr='phase.p.fluid_flow')
'''
Requirements that a simulation phase enable fluid flow and hence at least
intracellular fluid flow.

Simulation phases satisfying this requirement do *not* necessarily enable
extracellular fluid flow, which additionally require extracellular spaces to be
enabled.
'''


PRESSURE_OSMOTIC = SOLVER_FULL | phasereqsmake.make_requirements_bool_expr(
    name='osmotic pressure', bool_expr='phase.p.deform_osmo')
'''
Requirements that a simulation phase enable osmotic pressure.
'''


VOLTAGE_MEMBRANE_GHK = SOLVER_FULL | phasereqsmake.make_requirements_bool_expr(
    name='Goldman-Hodgkin-Katz (GHK) calculation',
    bool_expr='phase.p.GHK_calc')
'''
Requirements that a simulation phase enable alternative calculation of
transmembrane voltages (Vmem) given the Goldman-Hodgkin-Katz (GHK) equation.
'''

# ....................{ REQUIREMENTS ~ solver : ion       }....................
# Requirements satisfied by enabling only the full solver and a
# requirement-specific ion. Note that the requirement for the full solver is
# guaranteed by the phasereqsmake.make_requirements_ion() factory function.

ION_CALCIUM = phasereqsmake.make_requirements_ion(
    name='calcium ions (Ca2+)', ion_name='Ca')
'''
Requirements that a simulation phase enable calcium ions (Ca2+).
'''


ION_CHLORIDE = phasereqsmake.make_requirements_ion(
    name='chloride ions (Cl-)', ion_name='Cl')
'''
Requirements that a simulation phase enable chloride ions (Cl-).
'''


ION_POTASSIUM = phasereqsmake.make_requirements_ion(
    name='potassium ions (K+)', ion_name='K')
'''
Requirements that a simulation phase enable potassium ions (K+).
'''


ION_M_ANION = phasereqsmake.make_requirements_ion(
    name='M anions (M-)', ion_name='M')
'''
Requirements that a simulation phase enable M anions (M-).
'''


ION_SODIUM = phasereqsmake.make_requirements_ion(
    name='sodium ions (Na+)', ion_name='Na')
'''
Requirements that a simulation phase enable sodium ions (Na+).
'''

# ....................{ REQUIREMENTS ~ solver : ion : bool}....................
# Requirements satisfied by enabling only the full solver, a
# requirement-specific ion, and an additional requirement-specific boolean.

ION_CALCIUM_DYNAMICS = ION_CALCIUM | phasereqsmake.make_requirements_bool_expr(
    name='calcium ion (Ca2+) dynamics', bool_expr='phase.p.Ca_dyn')
'''
Requirements that a simulation phase enable calcium ions (Ca2+), complete
calcium dynamics, *and* a solver capable of fully simulating these features
(e.g., the full BETSE solver).
'''

# ....................{ REQUIREMENTS ~ solver : lambda    }....................
# Requirements satisfied by enabling only the full solver and a pair of
# requirement-specific lambda functions.

PRESSURE_TOTAL = SOLVER_FULL | phasereqsmake.make_requirements_funcs(
    name='total pressure',
    is_satisfied=lambda phase:
        phase.p.deform_osmo or phase.p.scheduled_options['pressure'] != 0,

    # For simplicity, this requirement is defined to be settable by enabling
    # osmotic pressure. While the mechanical pressure event could also be
    # enabled, doing so is less trivial than the former.
    set_satisfied=lambda phase: phase.p.__setattr__('deform_osmo', True))
'''
Requirements that a simulation phase enable at least one pressure feature:
namely, osmotic pressure or the mechanical pressure intervention.
'''


VOLTAGE_POLARITY = SOLVER_FULL | phasereqsmake.make_requirements_funcs(
    name='cellular voltage polarizability',
    is_satisfied =lambda phase: phase.p.cell_polarizability > 0,
    set_satisfied=lambda phase: phase.p.__setattr__(
        'cell_polarizability', 1e-4))
'''
Requirements that a simulation phase enable cellular voltage polarizability.
'''

# ....................{ REQUIREMENTS ~ solver : ecm       }....................
# Requirements satisfied by enabling only the full solver, extracellular
# spaces, and a single arbitrary requirement.

FLUID_EXTRA = FLUID | ECM
'''
Requirements that a simulation phase enable extracellular fluid flow and hence
fluid flow, extracellular spaces, *and* a solver capable of fully simulating
these features (e.g., the full BETSE solver).
'''


ION_CALCIUM_EXTRA = ION_CALCIUM | ECM
'''
Requirements that a simulation phase enable extracellular calcium ions (Ca2+)
and hence calcium ions, extracellular spaces, *and* a solver capable of fully
simulating these features (e.g., the full BETSE solver).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Facilities guaranteeing backward compatibility with prior file formats for
gene regulatory networks (GRNs).
'''

# ....................{ IMPORTS                            }....................
from betse.util.io.log import logs
from betse.util.type.types import type_check, MappingType



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **context manager** (i.e., classes implementing the context manager
protocol and hence the ``__enter__`` and ``__exit__`` special methods)
facilities.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Iterator
from contextlib import contextmanager

# ....................{ MANAGERS                           }....................
@contextmanager
def noop_context() -> Iterator[None]:
    '''
    Empty context manager exiting immediately after being entered.
    '''

    yield



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Primitive two-dimensional **convex polygon** (i.e., non-self-intersecting
polygon such that all line segments between any two points on the polygon
boundary remain strictly inside the polygon) functionality.
'''

#FIXME: Implement clipping algorithms that support concave clip polygons, which
#the Sutherland-Hodgman algorithm implemented below explicitly does *NOT*.
#Possible candidate replacements include:
#
#    https://en.wikipedia.org/wiki/Vatti_clipping_algorithm
#    https://en.wikipedia.org/wiki/Greiner%E2%80%93Hormann_clipping_algorithm
#    https://en.wikipedia.org/wiki/Weiler%E2%80%93Atherton_clipping_algorithm
#
#Vatti appears to be the most general-purpose solution and hence should
#possibly be our first avenue of inquiry.

# ....................{ IMPORTS                           }....................
from betse.util.io.log import logs
from betse.util.math.geometry import geopoint
from betse.util.math.geometry.polygon import geopoly
from betse.util.type.iterable import iterables
from betse.util.type.types import type_check, SequenceTypes

# ....................{ TESTERS                           }....................
#FIXME: Implement us up. To do so:
#* Choose any three adjacent polygon vertices -- say, "polygon[0]",
#  "polygon[1]", and "polygon[3]".
#* For brevity, call these points A, B, and C.
#* Then return True if and only if:
#  (Bx - Ax)(Cy - Ay) - (Cx - Ax)(By - Ay) > 0, which is equivalent to:
#  (Bx - Ax)(Cy - Ay) > (Cx - Ax)(By - Ay)
#For a partial derivation, see:
#    https://en.wikipedia.org/wiki/Curve_orientation#Practical_considerations
#FIXME: Unit test us up.
# @type_check
# def is_oriented_counterclockwise(polygon: SequenceTypes) -> bool:
#     '''
#     ``True`` only if the passed convex polygon is **oriented positively** (i.e.,
#     the vertices of this polygon are specified in counter-clockwise order).
#
#     Parameters
#     ----------
#     polygon : SequenceTypes
#         Two-dimensional sequence of all points defining the convex
#         two-dimensional polygon to be tested such that:
#         * The first dimension indexes each such point (in arbitrary order).
#         * The second dimension indexes each coordinate of this point such that:
#           * The first item is the X coordinate of this point.
#           * The second item is the Y coordinate of this point.
#
#     Returns
#     ----------
#     bool
#         ``True`` only if this convex polygon is oriented positively.
#     '''
#
#     pass

# ....................{ CLIPPERS                          }....................
@type_check
def clip(
    subject_polygon: SequenceTypes, clip_polygon: SequenceTypes,
) -> SequenceTypes:
    '''
    Clip the passed possibly non-convex subject polygon to the interior of the
    passed convex clip polygon by retracting all points of the former residing
    in the exterior of the latter onto the boundary of the latter.

    This function implements the Sutherland-Hodgman polygon clipping algorithm.
    Unlike most brute-force implementations of this algorithm, this
    implementation does *not* mandate the vertices of either polygon to be
    oriented (i.e., sorted) in either clockwise or counter-clockwise order.

    Parameters
    ----------
    subject_polygon : SequenceTypes
        Two-dimensional sequence of all points defining the **subject polygon**
        (i.e., possibly non-convex two-dimensional polygon to be clipped) such
        that:

        * The first dimension indexes each such point (in arbitrary order).
        * The second dimension indexes each coordinate of this point such that:

          * The first item is the X coordinate of this point.
          * The second item is the Y coordinate of this point.

        Note that this function expects the type of this sequence to define an
        ``__init__()`` method accepting a passed list as its first and only
        positional argument. Unsurprisingly, all builtin sequences (e.g.,
        :class:`tuple`, :class:`list`) *and* :mod:`numpy` sequences (e.g.,
        :class:`numpy.ndarray`) satisfy this requirement.
    clip_polygon : SequenceTypes
        Two-dimensional sequence of all points defining the **clip polygon**
        (i.e., strictly convex two-dimensional polygon to clip the subject
        polygon against) such that:

        * The first dimension indexes each such point (in arbitrary order).
        * The second dimension indexes each coordinate of this point such that:

          * The first item is the X coordinate of this point.
          * The second item is the Y coordinate of this point.

    Returns
    ----------
    SequenceTypes
        Copy of this subject polygon clipped to this clip polygon. The type of
        this sequence is the same as that of this subject polygon.
    '''

    # Orient the vertices of these polygons counter-clockwise.
    subject_polygon = geopoly.orient_counterclockwise(subject_polygon)
    clip_polygon    = geopoly.orient_counterclockwise(clip_polygon)

    # Clip this oriented subject polygon to this oriented clip polygon.
    return clip_counterclockwise(subject_polygon, clip_polygon)


@type_check
def clip_counterclockwise(
    subject_polygon: SequenceTypes, clip_polygon: SequenceTypes,
) -> SequenceTypes:
    '''
    Clip the passed possibly non-convex subject polygon to the interior of the
    passed convex clip polygon by retracting all points of the former residing
    in the exterior of the latter onto the boundary of the latter.

    This function implements the Sutherland-Hodgman polygon clipping algorithm.
    Like most brute-force implementations of this algorithm, this
    implementation mandates the vertices of both polygons to be oriented (i.e.,
    sorted) in strict counter-clockwise order. If the caller cannot guarantee
    this to be the case, the general-purpose but less efficient :func:`clip`
    function should be called instead. According to the GIGO (Garbage In,
    Garbage Out) principle, if the caller passes improperly oriented polygons,
    this function returns an improperly clipped polygon.

    See Also
    ----------
    :func:`clip`
        Further details on function signature (e.g., parameters, return value).
    https://en.wikipedia.org/wiki/Sutherland%E2%80%93Hodgman_algorithm
        Pseudocode outlining this algorithm.
    '''

    # If any passed sequence is *NOT* a polygon, raise an exception.
    geopoly.die_unless_polygon(subject_polygon, clip_polygon)

    #FIXME: If this clip polygon is *NOT* convex, raise an exception.

    #FIXME: Uncomment us up after implemented. Sadly, a minor issue remains:
    #while this will behave as expected for the convex clip polygon, this will
    #*NOT* behave as expected for the possibly non-convex subject polygon.
    #Detecting orientation in the latter case requires searching for the subject
    #polygon vertex with the smallest X and Y coordinates. *sigh*
    #die_unless_oriented_counterclockwise(subject_polygon, clip_polygon)

    # Current version of the clipped subject polygon to be returned, where
    # current implies the version of this subject polygon to be used in the
    # current iteration of the inner loop below.
    subj_poly_curr = subject_polygon

    # Previous version of the clipped subject polygon to be returned, where
    # previous implies the version of this subject polygon used in the prior
    # iteration of the inner loop below.
    subj_poly_prev = None

    # Current vertex of the clip polygon to clip against. The line segment
    # between this and the next such vertex is the current edge of this clip
    # polygon to clip the current edge of this subject polygon against.
    clip_vert_curr = clip_polygon[-1]

    # Next vertex of the clip polygon to clip against.
    clip_vert_next = None

    # Current vertex of the subject polygon to be clipped. The line segment
    # between this and the next such vertex is the current edge of this subject
    # polygon to be clipped against the current edge of this clip polygon.
    subj_vert_curr = None

    # Next vertex of the subject polygon to be clipped.
    subj_vert_next = None

    # True only if the current subject vertex resides inside this clip polygon.
    is_subj_vert_curr_inside = None

    # True only if the next subject vertex resides inside this clip polygon.
    is_subj_vert_next_inside = None

    # For each next vertex of the clip polygon to clip against...
    for clip_vert_next in clip_polygon:
        # If the previous iteration of the inner loop clipped the last three
        # vertices from the subject polygon, then the subject polygon resides
        # entirely outside of the clip polygon. In this case, the clipped
        # subject polygon is empty and no further work remains to be done.
        if len(subj_poly_curr) == 0:
            break

        # Preserve the previous version of the clipped subject polygon produced
        # by the previous iteration of the inner loop below *BEFORE* destroying
        # this sequence.
        subj_poly_prev = subj_poly_curr

        # Current version of the clipped subject polygon to be returned,
        # initialized to the empty list. For each edge of the previous version
        # of the clipped subject polygon at least partially residing inside the
        # clip polygon, the next iteration of the inner loop below appends at
        # least one vertex along this edge to this sequence.
        subj_poly_curr = []

        # Current vertex of the previous version of the clipped subject polygon
        # to be clipped, initialized to the last vertex of this polygon.
        subj_vert_curr = subj_poly_prev[-1]

        # This vertex resides inside this clip polygon if and only if this
        # vertex is spatially situated to the left of the vector signifying the
        # current edge of this clip polygon oriented in the counter-clockwise
        # orientation of this clip polygon, whose:
        #
        # * Head is the next vertex of this clip polygon.
        # * Tail is the current vertex of this clip polygon.
        is_subj_vert_curr_inside = geopoint.is_left_of_vector(
            subject_point=subj_vert_curr,
            vector_head_point=clip_vert_next,
            vector_tail_point=clip_vert_curr,)

        # For each next vertex of the previous version of the clipped subject
        # polygon...
        for subj_vert_next in subj_poly_prev:
            # This vertex resides inside this clip polygon if and only if this
            # vertex is spatially situated to the left of the same vector.
            is_subj_vert_next_inside = geopoint.is_left_of_vector(
                subject_point=subj_vert_next,
                vector_head_point=clip_vert_next,
                vector_tail_point=clip_vert_curr,)

            # If the current and next vertices of the previous version of the
            # clipped subject polygon reside on different sides of the current
            # edge of the clip polygon, the current edge of the former
            # necessarily intersects the current edge of the latter. In this
            # case, retract whichever of these two vertices resides outside of
            # the clip polygon to the boundary of the clip polygon.
            if is_subj_vert_curr_inside != is_subj_vert_next_inside:
                subj_poly_curr.append(geopoint.intersect_lines(
                    line1_point1=clip_vert_curr,
                    line1_point2=clip_vert_next,
                    line2_point1=subj_vert_curr,
                    line2_point2=subj_vert_next,
                ))

            # If the next vertex of the previous version of the clipped subject
            # polygon resides inside the current edge of the clip polygon,
            # preserve this vertex. To preserve the current orientation of the
            # subject polygon, this is performed *AFTER* possibly retracting a
            # subject vertex above.
            #
            # Note that similar logic need *NOT* be performed for the current
            # vertex of the previous version of the clipped subject polygon.
            # Why? Because if this vertex also resides inside the current edge
            # of the clip polygon, this vertex either:
            #
            # * Is the last vertex, in which case the last iteration of this
            #   loop is already guaranteed to preserve this vertex.
            # * Is any other vertex, in which case the prior iteration of this
            #   loop already preserved this vertex.
            #
            # In either case, this vertex is already preserved.
            if is_subj_vert_next_inside:
                subj_poly_curr.append(subj_vert_next)

            # Rotate the current vertex of the previous version of the clipped
            # subject polygon to the next such vertex.
            subj_vert_curr              = subj_vert_next
            is_subj_vert_curr_inside = is_subj_vert_next_inside

        # Rotate the current vertex of the clip polygon to the next such
        # vertex.
        clip_vert_curr = clip_vert_next

    # Return the current and hence final version of the clipped subject polygon
    # as a sequence of the same type as the subject polygon.
    return iterables.to_iterable(
        iterable=subj_poly_curr, cls=type(subject_polygon))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`pymod`-based decorators failing tests.

These decorators conditionally mark their decorated tests as failing depending
on whether the conditions signified by the passed parameters are satisfied
(e.g., the importability of the passed module name).
'''

# ....................{ IMPORTS                           }....................
import pytest
from betse.util.type.types import type_check

# ....................{ FAIL ~ alias                      }....................
xfail_if = pytest.mark.xfail
'''
Conditionally mark the decorated test as ignorably known to fail with the
passed human-readable justification if the passed boolean is ``False``.

Parameters
----------
boolean : bool
    Boolean to be tested.
reason : str
    Human-readable message justifying the failure of this test.

See Also
----------
:func:`pytest.mark.xfail`
    Further details on the ``XFAIL`` test state.
'''


@type_check
def xfail(reason: str):
    '''
    Unconditionally mark the decorated test as ignorably known to fail with the
    passed human-readable justification.

    :mod:`pytest` will neither run this test nor accumulate this failure. While
    superficially similar to tests unconditionally skipped via the ``@skip()``
    decorator, this failure will be collected as an `XFAIL` by py.test
    reporting.

    Parameters
    ----------
    reason : str
        Human-readable message justifying the failure of this test.
    '''

    return xfail_if(True, reason=reason)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-data descriptor** (i.e., objects satisfying the non-data
descriptor protocol, typically defined at class scope) facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassOrNoneTypes,
    ClassBoundMethodTypes,
)

# ....................{ CLASSES                            }....................
class ClassPropertyReadOnly(object):
    '''
    **Read-only class property** (i.e., property bound to a class rather than
    an instance of a class with no corresponding setter and hence read-only)
    descriptor.

    Decorators defined by this submodule (e.g., :func:`classproperty_readonly`)
    wrap their decorated callables with instances of this class.

    Caveats
    ----------
    While feasible, implementing writable class properties requires
    metaclasses, which invites subtle conflicts between uncooperative
    metaclasses in non-trivial class hierarchies. For maintainability, consider
    implementing writable class properties as customary class setters instead
    (e.g., ``def set_something(cls, something: object) -> None``).

    See also this `authoritative StackOverflow commentary`_ on the necessity of
    metaclasses in implementing writable class properties.

    .. _authoritative StackOverflow commentary:
       https://stackoverflow.com/questions/5189699/how-to-make-a-class-property

    Attributes
    ----------
    _getter : ClassBoundMethodTypes
        Class-bound getter method underlying this property.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, getter: ClassBoundMethodTypes) -> None:
        '''
        Initialize this read-only class property.

        Attributes
        ----------
        getter : ClassBoundMethodTypes
            Class-bound getter method underlying this property.
        '''

        # Classify this getter.
        self._getter = getter

    # ..................{ GETTERS                            }..................
    def __get__(self, obj: object, cls: ClassOrNoneTypes = None) -> object:
        '''
        Value of this read-only class property.
        '''

        # If this descriptor is accessed as an instance rather than class
        # variable, default to the class of this instance.
        if cls is None:
            cls = type(obj)

        # Return the result of calling the unbound function underlying the
        # underlying class-bound method descriptor with this object and class.
        return self._getter.__get__(obj, cls)()

# ....................{ DECORATORS                         }....................
@type_check
def classproperty_readonly(getter: CallableTypes) -> ClassPropertyReadOnly:
    '''
    **Read-only class property** (i.e., property bound to a class rather than
    an instance of a class with no corresponding setter and hence read-only),
    implemented by the passed unbound class method.

    Parameters
    ----------
    getter : CallableTypes
        **Unbound class method** (i.e., function whose first parameter is the
        subclass binding the property returned by this decorator).

    Returns
    ----------
    ClassPropertyReadOnly
        Read-only class property implemented by this unbound class method.

    See Also
    ----------
    :class:`ClassPropertyReadOnly`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objtest

    # If this getter is already a class-bound method wrapping an underlying
    # unbound function, raise an exception.
    objtest.die_if_instance(obj=getter, cls=ClassBoundMethodTypes)

    # Class method wrapping this getter.
    abstract_getter = classmethod(getter)

    # Return this method wrapped by a read-only class property descriptor.
    return ClassPropertyReadOnly(abstract_getter)


def abstractclassproperty_readonly(getter: CallableTypes) -> (
    ClassPropertyReadOnly):
    '''
    **Abstract read-only class property** (i.e., property bound to a class
    rather than an instance of a class with no corresponding setter and hence
    read-only, required to be reimplemented by subclasses), implemented by the
    passed unbound class method.

    See Also
    ----------
    :func:`classproperty_readonly`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.decorator.deccls import abstractclassmethod
    from betse.util.type.obj import objtest

    # If this getter is already a class-bound method wrapping an underlying
    # unbound function, raise an exception.
    objtest.die_if_instance(obj=getter, cls=ClassBoundMethodTypes)

    # Abstract class method wrapping this getter.
    abstract_getter = abstractclassmethod(getter)

    # Return this method wrapped by a read-only class property descriptor.
    return ClassPropertyReadOnly(abstract_getter)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level external process facilities.
'''

# ....................{ IMPORTS                           }....................
import sys
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ CONSTANTS                         }....................
SUCCESS = 0
'''
Exit status signifying a process to have terminated successfully.
'''


FAILURE_DEFAULT = 1
'''
Exit status typically signifying a process to have terminated prematurely with
a fatal error.

While any exit status in the range ``[1, 255]`` signifies failure, this exit
status is the most common and hence preferred default.
'''

# ....................{ EXCEPTIONS                        }....................
def raise_success() -> None:
    '''
    Raise the :class:`SystemExit` exception with the success exit status (i.e.,
    :data:`SUCCESS`), thus halting the current process with this status if
    uncaught.
    '''

    raise SystemExit(SUCCESS)


# ....................{ TESTERS                           }....................
@type_check
def is_success(exit_status: int) -> bool:
    '''
    ``True`` only if the passed exit status signifies success.
    '''

    return exit_status == SUCCESS


@type_check
def is_failure(exit_status: int) -> bool:
    '''
    ``True`` only if the passed exit status signifies failure.
    '''

    return exit_status != SUCCESS

# ....................{ EXITERS                           }....................
def exit_with_success(error_message: str = '') -> None:
    '''
    Halt the current process with the success exit status (i.e.,
    :data:`SUCCESS`), logging the passed error message if nonempty *or* exiting
    silently otherwise.
    '''

    exit_with_status(exit_status=FAILURE_DEFAULT, exit_message=error_message)


def exit_with_failure(error_message: str = '') -> None:
    '''
    Halt the current process with the default exit status for failure (i.e.,
    :data:`FAILURE_DEFAULT`), logging the passed error message if nonempty *or*
    exiting silently otherwise.
    '''

    exit_with_status(exit_status=FAILURE_DEFAULT, exit_message=error_message)


@type_check
def exit_with_status(
    exit_status: int = FAILURE_DEFAULT, exit_message: str = '') -> None:
    '''
    Halt the current process with the passed exit status, logging the passed
    message if nonempty *or* exiting silently otherwise.

    This message will be logged with level:

    * :attr:`LogLevel.INFO` if this exit status signifies success.
    * :attr:`LogLevel.ERROR` otherwise.
    '''

    # Log this message if nonempty.
    if exit_message:
        # If this status signifies success, log accordingly.
        if exit_status == SUCCESS:
            logs.log_info(exit_message)
        # Else, this status signifies failure. Log accordingly.
        else:
            logs.log_error(exit_message)

    # Exit the current process with this status.
    sys.exit(exit_status)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :func:`betse.util.type.cls.descriptors.expr_alias`
descriptor.
'''

# ....................{ IMPORTS                            }....................
import pytest
from enum import Enum
from pytest import fixture

# ....................{ SUPERCLASSES                       }....................
class ExprAliasBaseClass(object):
    '''
    Placeholder superclass of an arbitrary expression alias instantiated below.
    '''

    pass

# ....................{ GLOBALS                            }....................
_ElendilsOath = Enum('_ElendilsOath', (
    'SINOME', 'MARUVAN', 'AR', 'HILDINYAR', 'TENN' 'AMBAR', 'METTA',))
'''
Enumeration type to be aliased by the subsequent class. As required by the
:func:`expr_enum_alias` data descriptor, the names of all members of this
enumeration *must* be uppercase.
'''


_Namarie = Enum('_Namarie', (
    'NU', 'LUINI', 'YASSEN', 'TINTILAR', 'i', 'ELENI',))
'''
Enumeration type failing to satisfy the requirements of the
:func:`expr_enum_alias(`data descriptor. Specifically, at least one of the names
of a member of this enumeration is *not* uppercase.
'''

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def betse_expralias() -> object:
    '''
    Fixture creating and returning a mock object whose ad-hoc class contains
    instances of all expression alias descriptors to be unit tested.
    '''

    # Imports deferred for safety.
    from betse.util.type.descriptor.expralias import (
        expr_alias, expr_enum_alias)
    from betse.util.type.descriptor.datadescs import DataDescriptorBound

    # Class containing instances of this descriptor.
    class SongOfAragorn(object):
        # Descriptor aliasing an existing dictionary entry with class typing
        # and subclassing an ad-hoc base class.
        soa_line_four = expr_alias(
            expr='self._deep_roots["are not"]["reached by"]',
            cls=str,
            base_classes=(ExprAliasBaseClass,),
        )

        # Descriptor aliasing an existing dictionary entry with predicate
        # typing.
        soa_line_one = expr_alias(
            expr='self._deep_roots["All that is"]["gold does not"]',
            predicate=lambda value:
                value in ('glamour,', 'glimmer,','glitter,', 'glisten,',),
            predicate_label='adjectival',
        )

        # Descriptor aliasing an existing dictionary entry with both class and
        # predicate expression typing.
        soa_line_six = expr_alias(
            expr='self._deep_roots["A light from"]["the shadows shall"]',
            cls=str,
            predicate_expr=(
                "value in ('spring;', 'summer;', 'autumn;', 'winter;',)"),
            predicate_label='seasonal',
        )

        # Descriptor aliasing an existing dictionary entry without typing.
        soa_line_five = expr_alias(
            expr='self._deep_roots["From the ashes"]["a fire shall"]')

        # Descriptor aliasing an existing dictionary entry with enumerability,
        # whose corresponding value is defined below to be a lowercase string
        # whose uppercased equivalent is the name of an arbitrary member of the
        # enumeration defined above.
        soa_line_three = expr_enum_alias(
            expr='self._deep_roots["The old that is strong"]["does not wither"]',
            enum_type=_ElendilsOath)

        # Descriptor aliasing an existing dictionary entry with typing, whose
        # corresponding value is defined below to be of a different type safely
        # castable to the type declared here.
        or_line_three = expr_alias(
            expr='self._one_ring["Nine for Mortal Men"]["doomed to die,"]',
            cls=float)

        # Descriptor aliasing an existing dictionary entry with typing, whose
        # corresponding value is defined below to be of a different type *NOT*
        # safely castable to the type declared here.
        soa_line_two = expr_alias(
            expr='self._deep_roots["Not all those"]["who wander"]', cls=str)

        # Descriptor aliasing an existing dictionary entry with enumerability,
        # whose corresponding value is defined below to be of a non-enumeration
        # member.
        soa_line_seven = expr_enum_alias(
            expr='self._deep_roots["Renewed shall be"]["blade that was"]',
            enum_type=_ElendilsOath)

        # Descriptor aliasing a non-existing dictionary entry with typing.
        soa_line_none = expr_alias(
            expr='self._deep_roots["Many a man"]["his life hath"]', cls=str)


        def __init__(self) -> None:

            # Initialize the nested dictionary structures aliased above.
            self._deep_roots = {
                "are not": {
                    "reached by": "the frost.",
                },
                "All that is": {
                    "gold does not": "glitter,",
                },
                "Not all those": {
                    "who wander": 0xA3E01054,
                },
                "The old that is strong": {
                    "does not wither": _ElendilsOath.HILDINYAR.name.lower(),
                },
                "From the ashes": {
                    "a fire shall": "be woken,",
                },
                "A light from": {
                    "the shadows shall": "spring;",
                },
                "Renewed shall be": {
                    "blade that was": "broken,",
                },
                "The crownless again": {
                    "shall be": "king.",
                },
            }

            self._one_ring = {
                "Nine for Mortal Men": {
                    "doomed to die,": 1,
                },
            }

            # Alias an arbitrary entry of this dictionary for use below.
            the_crownless_again = self._deep_roots['The crownless again']

            # Object aliasing an existing dictionary entry with class typing.
            # For coverage, the optional "obj_name" parameter is also exercised.
            self.soa_line_eight = DataDescriptorBound(
                obj=the_crownless_again,
                data_desc=expr_alias(
                    expr='THE_CROWNLESS_AGAIN["shall be"]',
                    cls=str,
                    obj_name='THE_CROWNLESS_AGAIN',
                ))

    # Create and return an instance of this type.
    return SongOfAragorn()

# ....................{ TESTS                              }....................
def test_expralias_pass(betse_expralias) -> None:
    '''
    Test all aspects of the :func:`betse.util.type.cls.descriptors.expr_alias`
    descriptor intended to succeed.

    Parameters
    ----------
    betse_expralias : object
        Instance of this type to be tested.
    '''

    # Test the class-typed data descriptor's __get__() method.
    assert betse_expralias.soa_line_four == "the frost."

    # Test the numeric class-typed data descriptor's __get__() method,
    # implicitly casting a value of a different type to the desired type.
    assert type(betse_expralias.or_line_three) is float
    assert betse_expralias.or_line_three == 1.0

    # Test the numeric class-typed data descriptor's "expr_alias_cls" class
    # variable, accessed from this class rather than this instance.
    assert betse_expralias.__class__.or_line_three.expr_alias_cls is float

    # Test the predicate-typed data descriptor's __get__() method.
    assert betse_expralias.soa_line_one == "glitter,"

    # Test the class- and predicate-typed data descriptor's __get__()
    # method.
    assert betse_expralias.soa_line_six == "spring;"

    # Test the enumeration-typed data descriptor's __get__() method.
    assert betse_expralias.soa_line_three is _ElendilsOath.HILDINYAR

    # Test the untyped data descriptor's __get__() method.
    assert betse_expralias.soa_line_five == "be woken,"

    # Test the class-typed object's get() method.
    assert betse_expralias.soa_line_eight.get() == "king."

    # Test passing the class-typed data descriptor's __set__() method another
    # value of the same type.
    merchant_of_venice = "Gilded tombs do worms enfold."
    betse_expralias.soa_line_four = merchant_of_venice
    assert betse_expralias.soa_line_four == merchant_of_venice
    assert betse_expralias._deep_roots["are not"]["reached by"] == (
        merchant_of_venice)

    # Test passing the numeric class-typed data descriptor's __set__() method a
    # value of a different type safely castable to the desired type.
    betse_expralias.or_line_three = 9
    assert type(betse_expralias.or_line_three) is float
    assert betse_expralias.or_line_three == 9.0
    assert betse_expralias._one_ring[
        "Nine for Mortal Men"]["doomed to die,"] == 9.0

    # Test passing the predicate-typed data descriptor's __set__() method
    # another value of the same type.
    betse_expralias.soa_line_one              = 'glisten,'
    assert betse_expralias.soa_line_one      == 'glisten,'
    assert betse_expralias._deep_roots[
        "All that is"]["gold does not"] == ('glisten,')

    # Test passing the class- and predicate-typed data descriptor's __set__()
    # method another value of the same type.
    betse_expralias.soa_line_six                   = 'winter;'
    assert betse_expralias.soa_line_six           == 'winter;'
    assert betse_expralias._deep_roots[
        "A light from"]["the shadows shall"] == ('winter;')

    # Test passing the enumeration-typed data descriptor's __set__() method a
    # valid enumeration member.
    betse_expralias.soa_line_three = _ElendilsOath.MARUVAN
    assert betse_expralias.soa_line_three == _ElendilsOath.MARUVAN
    assert betse_expralias._deep_roots[
        "The old that is strong"]["does not wither"] == (
        _ElendilsOath.MARUVAN.name.lower())

    # Test passing the untyped data descriptor's __set__() method an arbitrary
    # non-string type.
    betse_expralias.soa_line_five = 0xCAFEBABE
    assert betse_expralias.soa_line_five == 0xCAFEBABE
    assert betse_expralias._deep_roots["From the ashes"]["a fire shall"] == (
        0xCAFEBABE)

    # Test passing the class-typed object's set() method another string value.
    prince_of_morocco = "Many a man his life hath sold"
    betse_expralias.soa_line_eight.set(prince_of_morocco)
    assert betse_expralias.soa_line_eight.get() == prince_of_morocco
    assert betse_expralias._deep_roots["The crownless again"]["shall be"] == (
        prince_of_morocco)

    # Test calling the class-typed data descriptor's __get__() implementation as
    # a class rather than instance variable.
    assert isinstance(betse_expralias.__class__.soa_line_four, ExprAliasBaseClass)


def test_expralias_fail(betse_expralias) -> None:
    '''
    Test all aspects of the :func:`betse.util.type.cls.descriptors.expr_alias`
    descriptor intended to fail.

    Parameters
    ----------
    betse_expralias : object
        Instance of this type to be tested.
    '''

    # Imports deferred for safety.
    from betse.exceptions import (
        BetseExprAliasException, BetseEnumException, BetseTypeException)
    from betse.util.type.descriptor.expralias import expr_enum_alias

    # Test instantiating the enumeration-typed data descriptor with an
    # enumeration type failing to satisfy this descriptor's requirements.
    with pytest.raises(BetseEnumException):
        expr_enum_alias(
            expr='wherein the stars tremble in the song of her voice,',
            enum_type=_Namarie)

    # Test passing the class-typed data descriptor's __set__() implementation a
    # non-string value.
    with pytest.raises(BetseTypeException):
        betse_expralias.soa_line_four = 0xFEEDBABE

    # Test passing the predicate-typed data descriptor's __set__()
    # implementation an unrecognized string value.
    with pytest.raises(BetseTypeException):
        betse_expralias.soa_line_one = 'expurgate,'

    # Test passing the class- and -predicate-typed data descriptor's __set__()
    # implementation both a non-string value and an unrecognized string value.
    with pytest.raises(BetseTypeException):
        betse_expralias.soa_line_six = 0xFEEDFACE
    with pytest.raises(BetseTypeException):
        betse_expralias.soa_line_six = 'extenuate;'

    # Test passing the enumeration-typed data descriptor's __set__()
    # implementation a non-enumeration member.
    with pytest.raises(BetseEnumException):
        betse_expralias.soa_line_three = _Namarie.TINTILAR

    # Test an invalid class-typed data descriptor's __get__() implementation,
    # aliased to access an existing key of an existing dictionary whose value
    # has an unexpected type.
    with pytest.raises(BetseTypeException):
        betse_expralias.soa_line_two

    # Test an invalid enumeration-typed data descriptor's __get__()
    # implementation, aliased to access an existing key of an existing
    # dictionary whose value is *NOT* the lowercase name of an enumeration
    # member.
    with pytest.raises(BetseEnumException):
        betse_expralias.soa_line_seven

    # Test an invalid untyped data descriptor's __get__() implementation,
    # aliased to access a non-existing key of an existing dictionary.
    with pytest.raises(BetseExprAliasException):
        betse_expralias.soa_line_none

    # Test the invalid data descriptor's __set__() implementation, aliased to
    # assign a non-existing key of an existing dictionary.
    with pytest.raises(KeyError):
        betse_expralias.soa_line_none = "Many a man his life hath sold"

    # Test calling an arbitrary data descriptor's __delete__() implementation,
    # which currently remains unimplemented to preserve backward compatibility.
    with pytest.raises(AttributeError):
        del betse_expralias.soa_line_four



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.


import numpy as np
from scipy.ndimage import gaussian_filter


def osmotic_P(sim, cells, p):

    # initialize osmotic pressures in cells and env
    sim.osmo_P_cell = np.sum(p.R * sim.T * sim.cc_cells, axis=0)
    sim.osmo_P_env = np.sum(p.R * sim.T * sim.cc_env, axis=0)

    # add in other molecular substances to the net sum:
    add_sum_cell = np.zeros(sim.cdl)

    if p.is_ecm is False:
        add_sum_env = np.zeros(sim.mdl)
    else:
        add_sum_env = np.zeros(sim.edl)

    if p.molecules_enabled:
        for nmol, mol in sim.molecules.core.molecules.items():
            add_sum_cell += mol.c_cells*p.R*sim.T
            add_sum_env += mol.c_env*p.R*sim.T

    if p.grn_enabled:
        for nmol, mol in sim.grn.core.molecules.items():
            add_sum_cell += mol.c_cells*p.R*sim.T
            add_sum_env += mol.c_env*p.R*sim.T

    # Add in the additional molecular concentrations to the main arrays:
    sim.osmo_P_cell += add_sum_cell
    sim.osmo_P_env += add_sum_env

    if p.is_ecm is False:

        op_env = np.dot(cells.M_sum_mems, sim.osmo_P_env)/cells.num_mems

        sim.osmo_P_delta = op_env - sim.osmo_P_cell

    else:

        sim.osmo_P_delta = sim.osmo_P_env[cells.map_cell2ecm] - sim.osmo_P_cell


    # Calculate the transmembrane flow of water due to osmotic pressure.
    # High, positive osmotic pressure leads to water flow into the cell. Existing pressure in the cell
    # resists the degree of osmotic influx. The effect also depends on aquaporin fraction in membrane:
    # sim.u_osmo = (sim.osmo_P_delta - sim.P_cells) * (p.aquaporins / p.mu_water)*cells.cell_sa
    sim.u_osmo = (sim.osmo_P_delta - sim.P_cells)*(p.aquaporins/p.mu_water)*(3e-10**2)*(1/p.tm)

    # obtain the divergence of the flow -- this is a strain rate:
    sim.div_u_osmo = sim.u_osmo*(cells.cell_sa/cells.cell_vol)

    # calculate pressure in whole network resulting from divergence :
    sim.PP = np.dot(cells.lapGJinv, -sim.div_u_osmo * p.rho * p.dt)

    # ------------------------------------------------------------------------------------------------------------
    # # actual volume change is amount of flow over the cell surface area per unit time:
    # sim.delta_vol = p.dt*sim.u_osmo*cells.cell_sa
    #
    # vol_ratio = cells.cell_vol / (cells.cell_vol - sim.delta_vol)
    #
    # # new concentrations in cells from C1*V1 = C2*V2 expression:
    # sim.cc_cells = sim.cc_cells * vol_ratio
    #
    # # reassign cell volume:
    # cells.cell_vol = cells.cell_vol - sim.delta_vol
    #
    # # reassign mem volume:
    # cells.mem_vol = cells.mem_vol/vol_ratio[cells.mem_to_cells]

    # if p.is_ecm is True:
    #     vo_ecm = cells.ecm_vol[cells.map_cell2ecm]
    #     v1_ecm = (1 - sim.delta_vol) * vo_ecm
    #
    #     for i, cc_array in enumerate(sim.cc_env):
    #         sim.cc_env[i][cells.map_cell2ecm] = cc_array[cells.map_cell2ecm] * (vo_ecm / v1_ecm)
    #
    #     cells.ecm_vol[cells.map_cell2ecm] = v1_ecm
    #
    # if p.voltage_dye is True:
    #     sim.cDye_cell = sim.cDye_cell * vol_ratio
    #
    # if p.scheduled_options['IP3'] != 0 or p.Ca_dyn is True:
    #     sim.cIP3 = sim.cIP3 * (vo / v1)

# def get_mass_flux(sim, cells, p):
#     """
#     Sum up individual trans-membrane and
#     trans-gap junction ion fluxes to obtain the
#     net flow of mass into a cell.
#
#     Assumes that each ion travels with a hydration
#     shell of 6 water molecules.
#
#     """
#
#     # calculate mass flux across cell membranes:
#     mass_flux = np.zeros(len(cells.mem_i))
#
#     for flux_array, mm in zip(sim.fluxes_mem, sim.molar_mass):
#         m_flx = flux_array * (mm + 6 * 18.01e-3)  # flux x molar mass of ion x 6 water molecules at 18e-3 kg/mol
#
#         mass_flux = mass_flux + m_flx
#
#     return mass_flux

    # # total mass change in cell
    # mass_change = self.mass_flux*p.dt*cells.mem_sa
    # # sum the change over the membranes to get the total mass change of salts:
    # self.delta_m_salts = np.dot(cells.M_sum_mems,mass_change)


#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all YAML-backed lists and list item subclasses.
'''

# ....................{ IMPORTS                           }....................
from abc import abstractmethod
from betse.lib.yaml.abc.yamlabc import YamlABC
# from betse.util.io.log import logs
from betse.util.type.cls import classes
from betse.util.type.iterable import iterget
from betse.util.type.obj import objtest
from betse.util.type.types import (
    type_check, ClassType, MappingOrSequenceTypes, SequenceTypes)
from collections.abc import MutableSequence

# ....................{ SUPERCLASSES ~ list item          }....................
class YamlListItemABC(YamlABC):
    '''
    Abstract base class of all **YAML-backed list item** (i.e., configuration
    intended to be added to a parent :class:`YamlList` container) subclasses.

    Each instance of this superclass may technically encapsulate any valid YAML
    type (e.g., :class:`int`, :class:`str`) but typically encapsulates a YAML
    dictionary of related key-value pairs (e.g., settings configuring a single
    animation or tissue profile).
    '''

    # ..................{ MAKERS                            }..................
    @classmethod
    def make_list(cls, *args, **kwargs) -> (
        'betse.lib.yaml.abc.yamllistabc.YamlList'):
        '''
        Create and return a YAML-backed list containing only items of this
        subclass type, both loaded from and savable back to a YAML-formatted
        file.

        Parameters
        ----------
        All parameters are passed as is to the :meth:`YamlList.__init__`
        method.
        '''

        return YamlList(*args, item_type=cls, **kwargs)

    # ..................{ MAKERS ~ abstract                 }..................
    # Subclasses *MUST* implement the following abstract class methods.

    @classmethod
    @abstractmethod
    def make_default(
        cls, yaml_list: 'betse.lib.yaml.abc.yamllistabc.YamlList') -> (
        'betse.lib.yaml.abc.yamllistabc.YamlListItemABC'):
        '''
        Create and return an instance of this subclass encapsulating a new
        dictionary containing default configuration settings.

        This class method is principally intended to be called by the
        :meth:`YamlList.append_default` method, appending this instance to an
        existing parent list of similar instances.

        Parameters
        ----------
        yaml_list : YamlList
            Parent list instantiating this instance.
        '''

        pass

    # ..................{ MAKERS ~ concrete                 }..................
    @classmethod
    @type_check
    def _make_loaded(cls, conf: MappingOrSequenceTypes) -> (
        'betse.lib.yaml.abc.yamllistabc.YamlListItemABC'):
        '''
        Create and return an instance of this subclass encapsulating the passed
        YAML-backed container of all configuration settings required by this
        subclass.

        Specifically, this method (in order):

        #. Instantiates a new instance of this class.
        #. Passes this container unmodified to the :meth:`YamlListItemABC.load`
           method bound to this instance.
        #. Returns this loaded instance.

        Design
        ----------
        This class method is a convenience helper *only* intended to be called
        by subclass implementations of the :meth:`make_default` method. While
        technically optional, this helper reduces boilerplate duplication and
        is thus preferable to the manual approach.

        Parameters
        ----------
        conf : MappingOrSequenceTypes
            Low-level container of related configuration settings loaded from
            and saved back to a YAML-formatted configuration file.
        '''

        # Instance of this subclass to be returned.
        yaml_list_item = cls()

        # Load this instance from this container.
        yaml_list_item.load(conf=conf)

        # Return this loaded instance.
        return yaml_list_item

# ....................{ SUPERCLASSES ~ list               }....................
class YamlList(YamlABC, MutableSequence):
    '''
    YAML-backed list subconfiguration both loadable from *and* savable back
    into a YAML-formatted file.

    Each item of this list is a :class:`YamlListItemABC` instance encapsulating
    the corresponding YAML-backed list item. Each such item may technically
    encapsulate any valid YAML type (e.g., :class:`int`, :class:`str`) but
    typically encapsulates a YAML dictionary of related key-value pairs (e.g.,
    animation settings, tissue profile).

    Design
    ----------
    This list conforms to the :class:`MutableSequence` API and hence defines
    all public methods mandated by that API, including the familiar
    :meth:`append`, :meth:`clear`, :meth:`count`, :meth:`extend`,
    :meth:`index`, :meth:`insert`, :meth:`pop`, :meth:`register`,
    :meth:`remove`, and :meth:`reverse` methods.

    Attributes
    ----------
    _conf : SequenceTypes
        Low-level sequence of related configuration settings loaded from and
        saved back to a YAML-formatted configuration file if the :meth:`load`
        method has been called *or* ``None`` otherwise.
    _confs_wrap : list
        High-level list of each instance of the :attr:`_item_type` subclass
        encapsulating each dictionary in the low-level :attr:`_conf` sequence.
    _item_type : ClassType
        Subclass of the :class:`YamlListItemABC` abstract base class with
        which to instantiate each wrapper in the high-level :attr:`_confs_wrap`
        list encapsulating each dictionary in the low-level :attr:`_conf` list.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, item_type: ClassType) -> None:
        '''
        Initialize this low-level YAML-backed list.

        Parameters
        ----------
        item_type : ClassType
            Subclass of the :class:`YamlListItemABC` abstract base class
            with which to encapsulate each low-level YAML-backed list item in
            the passed ``confs`` list. For each such item, a new instance of
            this subclass is appended to the internal :attr:`_confs_wrap` list
            storing these instances.
        '''

        # Initialize our superclass.
        super().__init__()

        # If the passed type is *NOT* a high-level YAML-backed list item
        # subclass, raise an exception.
        classes.die_unless_subclass(
            subclass=item_type, superclass=YamlListItemABC)

        # Classify all passed parameters.
        self._item_type = item_type

        # Nullify all remaining parameters for safety.
        self._confs_wrap = None

    # ..................{ SUPERCLASS ~ YamlABC              }..................
    # Concrete methods provided by our YamlABC superclass.

    @type_check
    def load(self, conf: SequenceTypes) -> None:

        # Load our superclass with the passed sequence.
        super().load(conf=conf)

        # Initialize a high-level list of high-level wrappers.
        self._confs_wrap = []

        # For each low-level dictionary in this low-level sequence...
        for conf_yaml in conf:
            # High-level wrapper of the caller-specified type.
            conf_wrap = self._item_type()
            conf_wrap.load(conf=conf_yaml)

            # Wrap this dictionary with this wrapper.
            self._confs_wrap.append(conf_wrap)


    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Nullify this high-level list of high-level wrappers.
        self._confs_wrap = None

    # ..................{ SUPERCLASS ~ mutable : abstract   }..................
    # Abstract methods required by our MutableSequence superclass.

    def __len__(self):
        return len(self._confs_wrap)


    @type_check
    def __delitem__(self, index: int) -> None:
        del self._conf[index]
        del self._confs_wrap[index]


    @type_check
    def __getitem__(self, index: int) -> YamlListItemABC:
        return self._confs_wrap[index]


    @type_check
    def __setitem__(self, index: int, value: YamlListItemABC) -> None:
        '''
        Set the low-level YAML-backed list item with the passed 0-based index
        to the passed such item.
        '''

        # If the passed object is *NOT* an instance of the caller-defined type,
        # raise an exception. This type is specified at initialization time and
        # hence cannot be type-checked via a method annotation.
        objtest.die_unless_instance(obj=value, cls=self._item_type)

        # Set the high-level list item with this index to this object.
        self._confs_wrap[index] = value

        # Set the low-level list item with this index to this object's
        # underlying YAML-backed dictionary.
        self._conf[index] = value.conf


    @type_check
    def insert(self, index: int, value: YamlListItemABC) -> None:
        '''
        Insert the passed low-level YAML-backed list item immediately *before*
        the item with the passed 0-based index.
        '''

        # Raise an exception unless the passed object is an instance of the
        # desired API, which is specified at initialization time and hence
        # cannot be type checked above by a method annotation.
        objtest.die_unless_instance(obj=value, cls=self._item_type)

        # Insert this object *BEFORE* the high-level list item with this index.
        self._confs_wrap.insert(index, value)

        # Insert this object's underlying YAML-backed dictionary *BEFORE* the
        # low-level list item with this index.
        self._conf.insert(index, value.conf)

    # ..................{ SUPERCLASS ~ mutable : concrete   }..................
    # Concrete methods defined by our MutableSequence superclass and overridden
    # here for dubious reasons (e.g., premature optimization).

    def clear(self) -> None:
        '''
        Remove all YAML-backed list items from this list, reducing this list to
        to the empty list.
        '''

        # Clear both the low- and high-level sequences wrapped by this object.
        self._conf.clear()
        self._confs_wrap.clear()

    # ..................{ APPENDERS                         }..................
    def append_default(self) -> YamlListItemABC:
        '''
        Append a new low-level YAML-backed list item initialized to default
        values suitable for this list *and* return this item.

        Returns
        ----------
        YamlListItemABC
            Instance of the :attr:`_item_type` subclass appended to the
            high-level :attr:`_confs_wrap` list, encapsulating the new list
            item appended to the low-level :attr:`_conf` list.
        '''

        # New simulation configuration list item specific to this list.
        conf_wrap = self._item_type.make_default(yaml_list=self)

        # Append this list item to this list.
        self.append(conf_wrap)

        # Return this list item.
        return conf_wrap

    # ..................{ GETTERS                           }..................
    @type_check
    def get_item_name_uniquified(self, item_name_format: str) -> str:
        '''
        Create and return a new **uniquified list item name** (i.e.,
        machine-readable string formatted from the passed format specifier,
        guaranteed to be unique across the YAML-backed ``name`` alias of each
        YAML-backed list item of this list).

        This method is typically called be subclass implementations of the
        :class:`YamlListItemABC.make_default` class method to enforce
        uniqueness of the default name of the newly created list item.

        Parameters
        ----------
        item_name_format : str
            Format specifier containing a ``{}`` substring (e.g.,
            ``Item ({}).``), interpolated by this function with an arbitrary
            integer to create the returned string.

        Returns
        ----------
        str
            Uniquified list item name formatted from this format specifier.

        Raises
        ----------
        BetseStrException
            If the passed format specifier contains no ``{}`` substring.
        '''

        # By hook or by crook, we will.
        return iterget.get_item_str_uniquified(
            iterable=self,
            item_attr_name='name',
            item_str_format=item_name_format,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Post-simulation animation** (i.e., animation produced *after* rather than
*while* solving a simulation) subclasses.
'''

# ....................{ IMPORTS                           }....................
from betse.science.config.export.visual.confexpvisabc import SimConfVisualCellsABC
from betse.science.phase.phasecls import SimPhase
from betse.science.visual.anim.animabc import AnimCellsABC
from betse.util.type.types import type_check, SequenceTypes

# ....................{ SUBCLASSES                        }....................
class AnimCellsAfterSolving(AnimCellsABC):
    '''
    Abstract base class of all post-simulation animation subclasses.

    Each subclass of this class animates arbitrary cell data as a time series
    plotted over the cell cluster (e.g., cell membrane voltage as a function of
    time) *after* rather than *while* solving simulations.
    '''

    @type_check
    def __init__(
        self,
        phase: SimPhase,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this post-simulation animation.

        Parameters
        ----------
        phase: SimPhase
            Current simulation phase.

        All remaining parameters are passed as is to the superclass
        ``__init__`` method.
        '''

        # Initialize our superclass.
        super().__init__(
            *args,

            # Pass this simulation phase as is to our superclass.
            phase=phase,

            # Save and show this post-simulation animation only if this
            # configuration enables doing so.
            is_save=phase.p.anim.is_after_sim_save,
            is_show=phase.p.anim.is_after_sim_show,

            # Save all post-simulation animations to the same parent directory.
            save_dir_parent_basename='anim',

            # Pass all remaining arguments as is to our superclass.
            **kwargs
        )


#FIXME: Merge this class into the "AnimCellsAfterSolving" superclass *AFTER*
#refactoring all subclasses to leverage layers.
class AnimCellsAfterSolvingLayered(AnimCellsAfterSolving):
    '''
    Post-simulation animation of arbitrary cell data as a time series (e.g.,
    cell membrane voltage as a function of time), plotted over the cell
    cluster using a predefined sequence of layers.
    '''

    # ..................{ SUPERCLASS                        }..................
    @type_check
    def __init__(
        self,
        conf: SimConfVisualCellsABC,
        layers: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        conf : SimConfVisualCellsABC
            Configuration for this visual, synchronized with the user-defined
            YAML-backed simulation configuration file for this phase.
        layers : SequenceTypes
            Sequence of all :class:`LayerCellsABC` instances collectively
            plotting each frame of this animation. See the
            :meth:`VisualCellsABC.__init__` method for further details.

        All remaining parameters are passed as is to the superclass
        :meth:`AnimCellsAfterSolving.__init__` method.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(
            *args,
            conf=conf,
            layers=layers,

            # Save this visual to a path containing this visual's name. Since
            # this name is guaranteed to be unique across all other visuals of
            # the same type, this path is guaranteed to be unique in at least
            # the directory containing all visuals of the same type.
            kind=conf.kind,
            **kwargs
        )

        # Display and/or save this animation.
        self._animate()

# ....................{ SUBCLASSES ~ obsolete             }....................
#FIXME: Replace use of all the following subclasses with the
#"AnimCellsAfterSolvingLayered" subclass *AFTER* refactoring all subclasses to
#leverage layers.
class AnimVelocity(AnimCellsAfterSolving):
    '''
    Abstract base class of all animations of a velocity flow plotted on the
    current cell cluster.
    '''

    def __init__(self, *args, **kwargs) -> None:

        # Pass all parameters *NOT* listed above to our superclass.
        super().__init__(
            *args,
            # Since this class already plots a streamplot, prevent the
            # superclass from plotting another streamplot as an overlay.
            is_current_overlayable=False,
            **kwargs
        )

        # Prefer an alternative colormap.
        self._colormap = self._phase.p.background_cm



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising functionality defined by the
:mod:`betse.util.app.meta.appmetamod` submodule.
'''

# ....................{ IMPORTS                           }....................
# import pytest
from betse.metadata import VERSION
from betse.util.test.pytest.mark.pytskip import skip_unless_requirement

# ....................{ TESTS                             }....................
# Exercising this function requires at least two modules satisfying the
# "betse.metadata" API. Since the only other module currently doing so is the
# "betsee.guimetadata" submodule, this test requires a recent BETSEE version
# guaranteed to be compatible with the current BETSE version. Since this exact
# merger is the principle use case for this function, exercising this use case
# substantially improves the reliability of BETSE <-> BETSEE integration.
@skip_unless_requirement('BETSEE >= {}.0'.format(VERSION))
def test_merge_module_metadeps() -> None:
    '''
    Unit test the :func:`betse.util.app.meta.appmetamod.merge_module_metadeps`
    function.
    '''

    # Defer heavyweight imports.
    from betse import metadeps as betse_metadeps
    from betse.util.app.meta import appmetamod
    from betse.util.app.meta.appmetamod import MERGE_MODULE_METADEPS_DICTS_NAME
    from betse.util.type.iterable.mapping import maptest
    from betse.util.type.obj import objtest
    from betsee import guimetadeps as betsee_metadeps

    # Application dependency metadata module merging the BETSE and BETSEE
    # application dependency metadata modules.
    module_metadeps_merged = appmetamod.merge_module_metadeps(
        # Fully-qualified name of the module to be created. For simplicity, we
        # create a guaranteeably unique test-specific submodule of the
        # top-level "betse_test" package, which is guaranteed to exist.
        module_name='betse_test.__betse_plus_betsee_metadeps__',
        modules_metadeps=(betse_metadeps, betsee_metadeps,),
    )

    # Test whether this module defines all requisite dictionary globals.
    objtest.die_unless_has_attr(
        module_metadeps_merged, *MERGE_MODULE_METADEPS_DICTS_NAME)

    # Test whether this module defines the following mandatory runtime
    # dependencies:
    #
    # * "Numpy", guaranteed to be required by BETSE.
    # * "PySide2", guaranteed to be required by BETSEE.
    maptest.die_unless_has_keys(
        mapping=module_metadeps_merged.RUNTIME_MANDATORY,
        keys=('Numpy', 'PySide2',))

    # Test whether this module defines the following optional runtime
    # dependencies:
    #
    # * "networkx", guaranteed to be required by BETSE.
    # * "pyside2uic", guaranteed to be required by BETSEE.
    maptest.die_unless_has_keys(
        mapping=module_metadeps_merged.RUNTIME_OPTIONAL,
        keys=('networkx', 'pyside2uic',))

    # Test whether this module defines the following mandatory test-time
    # dependencies:
    #
    # * "pytest", guaranteed to be required by both BETSE and BETSEE.
    maptest.die_unless_has_keys(
        mapping=module_metadeps_merged.TESTING_MANDATORY,
        keys=('pytest',))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.path.dirs` submodule.
'''

# ....................{ IMPORTS                            }....................
from betse.util.test.pytest.mark.pytskip import (
    # skip_if_ci_gitlab,
    skip_if_os_macos,
    skip_if_os_windows_vanilla,
)

# ....................{ TESTS                              }....................
#FIXME: Reenable this test under macOS when time permits. For unknown reasons,
#this test currently fails with an assertion error resembling:
#    >       assert (
#                dirs.get_mtime_recursive_newest(dirname) ==
#                paths.get_mtime_nonrecursive(subsubfilename2)
#            )
#    E       AssertionError: assert 1653971645.6776903 == 1653971645.677683
#
#We suspect the macOS-specific HFS+ filesystem to be the underlying culprit.
#According to this StackOverflow post, HFS+ provided at most one-second
#resolution on path timestamps:
#    https://stackoverflow.com/a/943537/2809027
#
#The above assertion error suggests HFS+ now provides finer-grain nanosecond
#resolution on path timestamps, but with the critical caveat that nanosecond
#resolution is *NOT* strictly reliable (e.g., due to low-level implementation
#details in the HFS+ journaling design).
#
#As a real-world solution, we might consider improving *ALL* BETSE-specific
#getter functions retrieving path timestamps (e.g.,
#get_mtime_recursive_newest(), get_mtime_nonrecursive()) to:
#* Conditionally detect when they are running under macOS.
#* If so, explicitly truncate the timestamp to be returned to only a certain
#  number of fractional digits. The above output suggests that HFS+ can reliably
#  provide at most four fractional digits, for example.
@skip_if_os_macos()
def test_dirs_get_mtime_newest(
    betse_temp_dir: 'py._path.local.LocalPath') -> None:
    '''
    Unit test the :func:`betse.util.path.dirs.get_mtime_recursive_newest` and
    related :func:`betse.util.path.paths.get_mtime_recursive_newest` functions
    with a directory tree fabricated in the passed temporary directory.

    Parameters
    ----------
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to this test.
    '''

    # Defer heavyweight imports.
    from betse.util.path import dirs, paths

    # Absolute paths of a subdirectory and file with arbitrary basenames
    # residing in this temporary directory.
    subdirpath  = betse_temp_dir.join("Quarks_Bar")
    subfilepath = betse_temp_dir.join("Grill")

    # Absolute paths of a subdirectory and two files with arbitrary basenames
    # residing in the above subdirectory residing in this temporary directory.
    subsubdirpath   = subdirpath.join("Gaming_House")
    subsubfilepath1 = subdirpath.join("And")
    subsubfilepath2 = subdirpath.join("Holosuite_Arcade")

    # Create these subdirectories and files, ensuring (...get it, "ensuring"?)
    # that the last such path is that asserted to be the most recent below.
    subdirpath.ensure(dir=True)
    subfilepath.ensure(file=True)
    subsubdirpath.ensure(dir=True)
    subsubfilepath1.ensure(file=True)
    subsubfilepath2.ensure(file=True)

    # Relevant paths as strings rather than "LocalPath" objects.
    dirname = str(betse_temp_dir)
    subsubdirname = str(subsubdirpath)
    subsubfilename2 = str(subsubfilepath2)

    # Ensure the last path created above to be the most recent.
    assert (
        dirs.get_mtime_recursive_newest(dirname) ==
        paths.get_mtime_nonrecursive(subsubfilename2)
    )

    # Update the mtime of an arbitrary subdirectory.
    subsubdirpath.setmtime()

    # Ensure this subdirectory to now be the most recent.
    assert (
        dirs.get_mtime_recursive_newest(dirname) ==
        paths.get_mtime_nonrecursive(subsubdirname)
    )

    # Ensure this subdirectory to now be the most recent when queried through
    # the related paths.get_mtime_recursive_newest() function.
    assert (
        paths.get_mtime_recursive_newest(
            (dirname, subsubdirname, subsubfilename2)) ==
        paths.get_mtime_nonrecursive(subsubdirname)
    )


#FIXME: Resolve the GitLab-CI-specific issue described under "Caveats" below.
#Ideally, this and *ALL* tests should run as is under *ALL* CI hosts. (The
#underlying culprit probably a recent change to GitLab's internal management of
#git repositories - possibly governed by the ${GIT_STRATEGY} variable.)
# @skip_if_ci_gitlab()
@skip_if_os_windows_vanilla()
def test_packages_init() -> None:
    '''
    Unit test the :func:`betse.util.path.dirs.recurse_subdirnames` function by
    validating that all **non-data subdirectories** (i.e., subdirectories
    containing only pure-Python) of *all* top-level package directories of this
    project (i.e., :mod:`betse` and :mod:`betse_test`) contain the mandatory
    ``__init__.py`` special file.

    Caveats
    -------
    **This test is currently incompatible with Microsoft Windows.** Why? Windows
    filesystems are case-insensitive in such a way as to inhibit the sanity of
    this test. Note the differing case in these two local variables of this test
    under a recent GitHub Actions-based Windows test workflow:

        # Note the "Lib" versus "lib" subdirectory basenames in these variables.
        EXCLUDE_DIRNAMES = {
            'D:\\a\\betse\\betse\\.tox\\py310-coverage\\Lib\\site-packages\\betse\\data',
            'D:\\a\\betse\\betse\\betse_test\\_data'
        }
        package_subdirname = (
            'D:\\a\\betse\\betse\\.tox\\py310-coverage\\lib\\site-packages\\betse\\data')

    **This test is currently incompatible with GitLab-CI's host environment.**
    GitLab-CI recently introduced a ``${GIT_STRATEGY}`` environment variable
    governing the tracking of git repositories. Previously, GitLab-CI appears
    to have performed a full-fledged clone of repositories; GitLab-CI appears
    to now be performing incremental pulls of repositories, instead. While
    ostensibly optimal from the efficiency perspective, the latter approach is
    clearly suboptimal from the reproducibility perspective. Each additional
    pull of a repository introduces incremental artifacts into the local
    copy of that repository, whose contents then begin to erroneously diverge.

    Why? Since ``git`` tracks files rather than directories, incremental pulls
    fail to implicitly prune formerly non-empty directories that have since
    been rendered empty by the subsequent removal of all files previously
    contained within those directories.

    While typically innocuous, the existence of empty subdirectories triggers
    false negatives from otherwise working functional and unit tests. In this
    case, this unit test recursively validates that all application
    (sub)packages define a mandatory ``__init__.py`` script, thus failing on
    the first empty non-data subdirectory. Until the underlying host-specific
    issue is resolved, this test is necessarily ignored on this host.
    '''

    # Defer heavyweight imports.
    import betse, betse_test
    from betse.util.app.meta import appmetaone
    from betse.util.io.log import logs
    from betse.util.path import dirs, files, pathnames
    from betse.util.py.module import pymodule
    from betse.util.type.text.string import strs

    # Tuple of all top-level packages.
    PACKAGES = (betse, betse_test)

    # Tuple of the absolute dirnames of all top-level package directories.
    PACKAGE_DIRNAMES = (pymodule.get_dirname(package) for package in PACKAGES)

    # Set of the absolute dirnames of all top-level directories to be .
    EXCLUDE_DIRNAMES = {
        appmetaone.get_app_meta().data_dirname,
        appmetaone.get_app_meta().test_data_dirname,
    }

    # For each such dirname...
    for package_dirname in PACKAGE_DIRNAMES:
        # Log this inspection.
        logs.log_info('')
        logs.log_info(
            'Searching "%s/" for non-package subdirectories...',
            pathnames.get_basename(package_dirname))

        # For the absolute direname of each direct and transitive subdirectory
        # of this package directory...
        for package_subdirname in dirs.recurse_subdirnames(package_dirname):
            # If this is either (in decreasing order of test efficiency)...
            #
            if (
                # A cache subdirectory *OR*...
                pathnames.get_basename(package_subdirname) == '__pycache__' or
                # An empty subdirectory *OR*...
                dirs.is_empty(package_subdirname) or
                # Either an excludable directory or subdirectory of such
                # directory...
                any(
                    strs.is_prefix(
                        text=package_subdirname, prefix=exclude_dirname)
                    for exclude_dirname in EXCLUDE_DIRNAMES
                )
            # ...then this subdirectory is guaranteed to contain no subpackages
            # and hence be safely ignorable.
            ):
                # Log this exclusion.
                logs.log_info('Excluding subdirectory: %s', package_subdirname)

                # Continue to the next subdirectory.
                continue
            # Else, this subdirectory is expected to be a subpackage.

            # Log this inspection.
            logs.log_info('Including subdirectory: %s', package_subdirname)

            # If this subdirectory does *NOT* contain an "__init__.py" file,
            # raise an exception.
            files.join_or_die(package_subdirname, '__init__.py')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **Git** (i.e., third-party version control system remotely tracking
all changes to this application) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseGitException
# from betse.util.io.log.logenum import LogLevel
from betse.util.type.types import type_check, ModuleType, StrOrNoneTypes

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_worktree(dirname: str) -> None:
    '''
    Raise an exception unless the directory with the passed pathname is a **Git
    working tree** (i.e., contains a ``.git`` subdirectory).

    Parameters
    ----------
    dirname : str
        Relative or absolute pathname of the directory to be validated.

    Raises
    ----------
    BetseGitException
        If this directory is *not* a Git working tree.

    See Also
    ----------
    :func:`is_repo`
        Further details.
    '''

    # If this directory is *NOT* a Git working tree, raise an exception.
    if not is_worktree(dirname):
        raise BetseGitException(
            'Directory "{}" not a Git working tree '
            '(i.e., contains no ".git" subdirectory).'.format(dirname))

# ....................{ TESTERS                           }....................
@type_check
def is_worktree(dirname: str) -> bool:
    '''
    ``True`` only if the directory with the passed pathname is a **Git working
    tree** (i.e., contains a ``.git`` subdirectory).

    Parameters
    ----------
    dirname : str
        Relative or absolute pathname of the directory to be tested.

    Returns
    ----------
    bool
        ``True`` only if this directory is a Git working tree.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, pathnames

    # Absolute or relative pathname of this directory's ".git" subdirectory.
    git_subdirname = pathnames.join(dirname, '.git')

    # Return True only if this subdirectory exists.
    return dirs.is_dir(git_subdirname)

# ....................{ GETTERS                           }....................
@type_check
def get_package_worktree_dirname_or_none(
    package: ModuleType) -> StrOrNoneTypes:
    '''
    **Absolute canonical dirname** (i.e., absolute dirname after resolving
    symbolic links) of the **Git-based working tree** (i.e., top-level
    directory containing the canonical ``.git`` subdirectory) governing the
    passed top-level Python package if found *or* ``None`` otherwise.

    Caveats
    ----------
    **This directory typically does not exist.** This directory is only
    required during development by developers and hence should *not* be assumed
    to exist.

    For safety and efficiency, this function does *not* recursively search the
    filesystem up from the directory yielding this package for a ``.git``
    subdirectory; rather, this function only directly searches that directory
    itself for such a subdirectory. For that reason, this function typically
    returns ``None`` for *all* modules except top-level packages installed in a
    developer manner (e.g., by running ``sudo python3 setup.py develop``).

    Parameters
    ----------
    package : ModuleType
        Top-level Python package to be queried.

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * The absolute canonical dirname of that package's Git working tree if
          that package was installed in a developer manner: e.g., by

          * ``python3 setup.py develop``.
          * ``python3 setup.py symlink``.

        * ``None`` if that package was installed in a non-developer manner:
          e.g., by

          * ``pip3 install``.
          * ``python3 setup.py develop``.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, pathnames
    from betse.util.py.module import pymodule

    # Absolute canonical dirname of the directory providing this package,
    # canonicalized into a directory rather than symbolic link to increase the
    # likelihood of obtaining the actual parent directory of this package
    package_dirname = pymodule.get_dirname_canonical(package)

    # Absolute dirname of the parent directory of this directory.
    worktree_dirname = pathnames.get_dirname(package_dirname)

    # Absolute dirname of the ".git" subdirectory of this parent directory.
    git_subdirname = pathnames.join(worktree_dirname, '.git')

    # Return this parent directory's absolute dirname if this subdirectory
    # exists *OR* "None" otherwise.
    return worktree_dirname if dirs.is_dir(git_subdirname) else None

# ....................{ CLONERS                           }....................
@type_check
def clone_worktree_shallow(
    branch_or_tag_name: str,
    src_dirname: str,
    trg_dirname: str,
) -> None:
    '''
    **Shallowly clone** (i.e., recursively copy without commit history) from
    the passed branch or tag name of the source Git working tree with the
    passed dirname into the target non-existing directory.

    For space efficiency, the target directory will *not* be a Git working tree
    (i.e., will *not* contain a ``.git`` subdirectory). This directory will
    only contain the contents of the source Git working tree isolated at
    either:

    * The ``HEAD`` of the branch with the passed name.
    * The commit referenced by the tag with the passed name.

    Caveats
    ----------
    Due to long-standing issues in Git itself, the passed branch or tag name
    *cannot* be the general-purpose SHA-1 hash of an arbitrary commit. While
    performing a shallow clone of an arbitrary commit from such a hash is
    technically feasible, doing so requires running a pipeline of multiple
    sequential Git subprocesses and hence exceeds the mandate of this function.

    Parameters
    ----------
    branch_or_tag_name : str
        Name of the branch or tag to clone from this source Git working tree.
    src_dirname : str
        Relative or absolute pathname of the source Git working tree to clone
        from.
    trg_dirname : str
        Relative or absolute pathname of the target non-existing directory to
        clone into.

    Raises
    ----------
    BetseGitException
        If this source directory is *not* a Git working tree.
    BetseDirException
        If this target directory already exists.

    See Also
    ----------
    https://stackoverflow.com/questions/26135216/why-isnt-there-a-git-clone-specific-commit-option
        StackOverflow question entitled "Why Isn't There A Git Clone Specific
        Commit Option?", detailing Git's current omission of such
        functionality.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import paths, pathnames
    from betse.util.os.command import cmdrun

    # If this source directory is *NOT* a Git working tree, raise an exception.
    die_unless_worktree(src_dirname)

    # If this target directory already exists, raise an exception.
    paths.die_if_path(trg_dirname)

    # Absolute pathname of the source Git working tree to clone from. The
    # "git clone" command requires "file:///"-prefixed absolute pathnames.
    src_dirname = pathnames.canonicalize(src_dirname)

    # Git-specific URI of this source Git working tree.
    src_dir_uri = 'file://' + src_dirname

    # Tuple of shell words comprising the Git command performing this clone.
    git_command = (
        'git', 'clone',

        # Branch or tag to be cloned.
        '--branch', branch_or_tag_name,

        # Perform a shallow clone.
        '--depth', '1',

        # From this source to target directory.
        src_dir_uri, trg_dirname,
    )

    # Shallowly clone this source to target directory. Contrary to expectation,
    # "git" redirects non-error or -warning output resembling the following to
    # stderr rather than stdout:
    #
    #    Cloning into '/tmp/pytest-of-leycec/pytest-26/cli_sim_backward_compatibility0/betse_old'...
    #    Note: checking out 'd7d6bf6d61ff2b467f9983bc6395a8ba9d0f234e'.
    #
    #    You are in 'detached HEAD' state. You can look around, make experimental
    #    changes and commit them, and you can discard any commits you make in this
    #    state without impacting any branches by performing another checkout.
    #
    #    If you want to create a new branch to retain commits you create, you may
    #    do so (now or later) by using -b with the checkout command again. Example:
    #
    #      git checkout -b <new-branch-name>
    #
    # This is distasteful. Clearly, this output should *NOT* be logged as either
    # an error or warning. Instead, both stdout and stderr output would ideally
    # be logged as informational messages. Unfortunately, the following call
    # appears to erroneously squelch rather than redirect most stderr output:
    #
    #    cmdrun.log_output_or_die(
    #        command_words=git_command,
    #        stdout_log_level=LogLevel.INFO,
    #        stderr_log_level=LogLevel.ERROR,
    #    )
    #
    # Frankly, we have no idea what is happening here -- and it doesn't
    # particularly matter. The current approach, while non-ideal, suffices.
    cmdrun.run_or_die(command_words=git_command)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level classes aggregating all parameters pertaining to simulation events.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseSimEventException
from betse.science.tissue.event.tiseveabc import SimEventPulseABC
from betse.science.math import toolbox
from betse.science.phase.phasecls import SimPhase
# from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ SUBCLASSES                        }....................
class SimEventPulseVoltage(SimEventPulseABC):
    '''
    **Voltage event** (i.e., event applying a directed voltage to the
    environmental boundary for a range of simulation time steps).

    A positive voltage will be applied to one boundary edge during this period;
    a negative voltage will be applied to another.

    Attributes
    ----------------------------
    peak_voltage : float
        Maximum voltage (V) to be applied.
    positive_voltage_boundary : str
        Character identifying the boundary edge to apply this positive voltage
        to. Valid values include:

        * ``T``, the top boundary.
        * ``B``, the bottom boundary.
        * ``L``, the left boundary.
        * ``R``, the right boundary.
    negative_voltage_boundary : str
        Character identifying the boundary edge to apply this negative voltage
        to. Valid values are as for ``side_positive_voltage`` above.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize this voltage event for the passed simulation configuration.

        Attributes
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        aev = p._conf['apply external voltage']

        # Initialize our superclass with some passed parameters.
        super().__init__(
            p=p,
            start_time_step=float(aev['change start']),
            stop_time_step=float(aev['change finish']),
            time_step_rate=float(aev['change rate']),
        )

        # Classify all remaining parameters.
        self.peak_voltage = float(aev['peak voltage'])
        self.positive_voltage_boundary = _convert_boundary_str_to_char(
            aev['positive voltage boundary'])
        self.negative_voltage_boundary = _convert_boundary_str_to_char(
            aev['negative voltage boundary'])


    #FIXME: Refactor to resemble the superclass method signature.
    @type_check
    def fire(self, phase: SimPhase, time_step: float) -> None:

        effector = toolbox.pulse(
            time_step,
            self.start_time_step,
            self.stop_time_step,
            self.time_step_rate,
        )

        phase.sim.bound_V[self.positive_voltage_boundary] = (
             self.peak_voltage * effector)
        phase.sim.bound_V[self.negative_voltage_boundary] = (
            -self.peak_voltage * effector)

# ....................{ CONVERTERS                         }....................
#FIXME: Refactor this function to leverage a private global dictionary: e.g.,
#
#    BOUNDARY_STR_TO_CHAR = {
#        'top':    'T',
#        'bottom': 'B',
#        'left':   'L',
#        'right':  'R',
#    }
#    boundary_char = BOUNDARY_STR_TO_CHAR.get(side, None)
#    if boundary_char is None:
#        raise BetseSimConfException(
#            'Boundary edge "{}" unrecognized.'.format(side))
#    return boundary_char
#
#The sane means of doing so would be to refactor the implementations of the
#"positive_voltage_boundary" and "positive_voltage_boundary" attributes above
#into @property_cached-style properties. Possibly? Or not. That does rather
#seem like overkill. The current function-based approach is probably superior.
#FIXME: Actually, why do we even require character constants like "T" here?
#These should all be leveraging enumeration member constants, in which the
#conversion from human-readable names to enumeration member constants will
#already be implicitly defined by the "Enum" class -- which, in turn, will
#enable us to entirely remove this function.
@type_check
def _convert_boundary_str_to_char(side: str) -> str:
    '''
    Convert the passed human-readable string constant identifying an
    environmental boundary edge (e.g., `top`) into the corresponding
    machine-readable character constant (e.g., `T`).
    '''

    if   side == 'top':    return 'T'
    elif side == 'bottom': return 'B'
    elif side == 'left':   return 'L'
    elif side == 'right':  return 'R'
    else:
        raise BetseSimEventException(
            'Boundary edge "{}" unrecognized.'.format(side))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Matplotlib-specific classes writing animations as video.
'''

#FIXME: Consider contributing most or all of this submodule back to matplotlib.

# ....................{ IMPORTS                           }....................
from matplotlib.animation import FFMpegBase, MovieWriter, writers
from betse.exceptions import BetseMatplotlibException
from betse.util.io.log import logs
from betse.util.os.shell import shellstr
from betse.util.os.command import cmdrun
from betse.util.type.cls import classes
from betse.util.type.iterable.mapping.mapcls import ReversibleDict
from betse.util.type.text import regexes
from betse.util.type.text.string import strjoin
from betse.util.type.types import (
    type_check, ClassType, NoneType, SequenceTypes, StrOrNoneTypes)

# ....................{ DICTS                             }....................
WRITER_NAME_TO_COMMAND_BASENAME = ReversibleDict(
    # AVConv-based video encoding with pipe-based writing.
    avconv='avconv',

    # AVConv-based video encoding with file-based writing.
    avconv_file='avconv',

    # FFmpeg-based video encoding with pipe-based writing.
    ffmpeg='ffmpeg',

    # FFmpeg-based video encoding with file-based writing.
    ffmpeg_file='ffmpeg',

    # Mencoder-based video encoding with pipe-based writing.
    mencoder='mencoder',

    # Mencoder-based video encoding with file-based writing.
    mencoder_file='mencoder',

    # ImageMagick-based animated GIF encoding with pipe-based writing.
    imagemagick='convert',

    # ImageMagick-based animated GIF encoding with file-based writing.
    imagemagick_file='convert',
)
'''
Bidirectional dictionary mapping the matplotlib-specific name of each video
encoder supported by matplotlib (e.g., `imagemagick`) to and from the basename
of that encoder's external command (e.g., `convert`).

These mappings are accessible as follows:

* `WRITER_NAME_TO_COMMAND_BASENAME`, the forward mapping from encoder names to
  command basenames.
* `WRITER_NAME_TO_COMMAND_BASENAME.reverse`, the reverse mapping from command
  basenames to encoder names.
'''


# Subsequently initialized by the init() function.
WRITER_BASENAME_TO_CONTAINER_FILETYPE_TO_CODEC_NAMES = None
'''
Dictionary mapping from the basename of the external command for each video
encoder recognized by matplotlib (e.g., ``ffmpeg``) to a nested dictionary
mapping from the filetype of each video container format recognized by BETSE
(e.g., ``mp4``) to a list of the names of all widely used video codecs
supported by that encoder (in descending order of general-purpose, subjective
preference).

Since multiple matplotlib animation writers (e.g., ``ffmpeg``, ``ffmpeg_file``)
typically execute the same external command (e.g., ``ffmpeg``), this dictionary
maps from the latter rather than the former. Mapping from the former would
needlessly require redundant mappings for writers sharing the same command.

This dictionary is principally used by the
:func:`get_first_codec_name_supported` getter function to obtain the preferred
codec for a given combination of video writer and container format.
'''

# ....................{ INITIALIZERS                      }....................
# For simplicity, this function is called below on the first importation of
# this submodule rather than explicitly called by callers.
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    _init_globals()


def _init_globals() -> None:
    '''
    Initialize all global variables declared by this submodule.
    '''

    # Globals initialized below.
    global WRITER_BASENAME_TO_CONTAINER_FILETYPE_TO_CODEC_NAMES

    WRITER_BASENAME_TO_CONTAINER_FILETYPE_TO_CODEC_NAMES = {
        # FFmpeg.
        'ffmpeg': {
            # Audio Video Interleave (AVI). AVI supports the smallest subset of
            # MPEG-centric codecs of all recognized container formats and hence
            # serves as the baseline for listing such codecs.
            'avi': (
                # H.264 / AVC / MPEG-4 Part 10.
                'libx264',
                # MPEG-4 Part 2 Advanced Simple Profile (ASP) via an external
                # shared library, typically providing better quality at lower
                # bitrates than the otherwise equivalent built-in codec for
                # MPEG-4 Part 2 (i.e., "mpeg4").
                'libxvid',
                # MPEG-4 Part 2 Advanced Simple Profile (ASP) support built-in
                # to all FFmpeg installations.
                'mpeg4',
                # H.263. Assuming videos leveraging this codec to be encoded as
                # basic "baseline" H.263 bitstreams, these videos are decodable
                # as is by conventional MPEG-4 decoders. To quote the eponymous
                # Wikipedia article on H.263: "MPEG-4 Part 2 is H.263
                # compatible in the sense that basic 'baseline' H.263
                # bitstreams are correctly decoded by an MPEG-4 Video decoder."
                'h263',
                # MPEG-2.
                'mpeg2video',
            ),

            # Graphics Interchange Format (GIF).
            'gif': ('gif',),

            # Matroska.
            'mkv': None,

            # QuickTime.
            'mov': None,

            # MPEG-4 Part 14.
            'mp4': None,

            # Theora.
            'ogv': ('libtheora',),

            # WebM.
            'webm': (
                #FIXME: As of this writing (i.e., mid-2017), VP-9 is poorly
                #supported in media players (e.g., mpv, vlc) and hence
                #disabled. Reevaluate and consider enabling this otherwise
                #sensible default at some future date.
                # 'libvpx-vp9',  # WebM VP-9

                'libvpx',      # WebM VP-8
            ),
        },

        # Libav.
        'avconv': None,

        # Mencoder.
        'mencoder': None,

        # ImageMagick, encoding videos only as old-school animated GIFs. Since
        # ImageMagick is *NOT* a general-purpose video encoder and hence fails
        # to support the notion of video codecs, a codec of "None" is used.
        # This has beneficial side effects, including ensuring that no
        # edge-case functionality in the matplotlib codebase attempts to
        # erroneously encode anything with ImageMagick using a codec.
        'convert': {
            'gif': (None,),
        }
    }

    # Shorthand to preserve sanity below.
    codec_names = WRITER_BASENAME_TO_CONTAINER_FILETYPE_TO_CODEC_NAMES

    # For FFmpeg, define the set of all codecs supported by the "mp4" (i.e.,
    # MPEG-4 Part 14) and "mov" (i.e., QuickTime) container formats to be the
    # those supported by the now-obsolete AVI container format plus all newer
    # codecs *NOT* supported by that format.
    codec_names['ffmpeg']['mp4'] = (
        # H.264 / AVC / MPEG-4 Part 10.
        'libx264',
        # H.265 / HEVC / MPEG-H Part 2. Since H.265 is significantly less
        # well-supported than H.264, the former takes precedence.
        'hevc',
    # All older codecs supported by the now-obsolete AVI container format.
        # Since the first codec is the aforementioned "libx264", this codec
        # (and only this codec) is ignored with a simple slice.
    ) + codec_names['ffmpeg']['avi'][1:]
    codec_names['ffmpeg']['mov'] = codec_names['ffmpeg']['mp4']

    # For FFmpeg, define the set of all codecs supported by the "mkv" (i.e.,
    # Matroska) container format to the set of all codecs supported by all
    # other container formats (excluding animated GIFs), giving preference to
    # open-standards codecs rather than proprietary codecs. Matroska: rock it!
    codec_names['ffmpeg']['mkv'] = (
        codec_names['ffmpeg']['webm'] +
        codec_names['ffmpeg']['ogv'] +
        codec_names['ffmpeg']['mp4']
    )

    # Define Libav to support exactly all codecs supported by FFmpeg. Since the
    # two are well-synchronized forks of each other attempting (and mostly
    # succeeding) to preserve a common command-line API, assuming codec parity
    # is typically a safe assumption.
    codec_names['avconv'] = codec_names['ffmpeg']

    # Define Mencoder to support exactly all codecs supported by FFmpeg. Since
    # Mencoder internally leverages the same "libavcodec" shared library
    # leveraged by FFmpeg, assuming codec parity is *ALWAYS* a safe assumption.
    # While Mencoder also provides a small handful of Mencoder-specific codecs
    # (see "mencoder -ovc help"), these codecs are commonly regarded as
    # inferior to their "libavcodec" counterparts. In either case, matplotlib
    # internally mandates use of "libavcodec"-provided codecs rather than
    # Mencoder-specific codecs in all Mencoder writer classes (e.g.,
    # "MencoderWriter").
    codec_names['mencoder'] = codec_names['ffmpeg']

# ....................{ EXCEPTIONS                        }....................
def die_unless_writer(writer_name: str) -> None:
    '''
    Raise an exception unless a matplotlib animation writer class registered
    with the passed name is recognized by both BETSE and matplotlib, implying
    the external command required by this class to be installed on the current
    system.

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to validate.

    See Also
    ----------
    :func:`is_writer`
        Further details.
    '''

    # For human-readable granularity in exception messages, call granular
    # testers rather than the catch-all is_writer() tester.
    #
    # If this writer is unrecognized by BETSE, raise an exception.
    die_unless_writer_betse(writer_name)

    # If this writer is unrecognized by matplotlib...
    if not is_writer_mpl(writer_name):
        # Basename of this writer's command.
        writer_basename = WRITER_NAME_TO_COMMAND_BASENAME[writer_name]

        # Raise this exception.
        raise BetseMatplotlibException(
            'Matplotlib animation video writer "{}" '
            'not registered with matplotlib '
            '(i.e., command "{}" not found).'.format(
                writer_name, writer_basename))


def die_unless_writer_betse(writer_name: str) -> None:
    '''
    Raise an exception unless a matplotlib animation writer class registered
    with the passed name is recognized by BETSE.

    If an exception is *not* raised, this does *not* imply the external command
    required by this class to be runnable or even installed on this system.

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to validate.

    See Also
    ----------
    is_writer_betse
        Further details.
    '''

    if not is_writer_betse(writer_name):
        raise BetseMatplotlibException(
            'Matplotlib animation video writer "{}" '
            'unrecognized by BETSE.'.format(writer_name))

# ....................{ EXCEPTIONS ~ command              }....................
def die_unless_writer_command(writer_basename: str) -> None:
    '''
    Raise an exception unless at least one matplotlib animation writer class
    running the external command with the passed basename is recognized by both
    BETSE and matplotlib, implying this command to be installed on the current
    system.

    Parameters
    ----------
    writer_basename : str
        Basename of the external command of the writer to validate.

    See Also
    ----------
    is_writer_command
        Further details.
    '''

    if not is_writer_command(writer_basename):
        raise BetseMatplotlibException(
            'Matplotlib animation video writer command "{basename}" '
            'unrecognized by BETSE or not registered with matplotlib '
            '(i.e., command "{basename}" not found).'.format(
                basename=writer_basename))

# ....................{ TESTERS                           }....................
@type_check
def is_writer(writer_name: str) -> bool:
    '''
    ``True`` only if a matplotlib animation writer class (e.g.,
    :class:`FFMpegWriter`) registered with the passed name (e.g., ``ffmpeg``)
    is recognized by both BETSE and matplotlib, implying the external command
    required by this class (e.g., ``fmpeg``) to be installed on the current
    system.

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to test.

    Returns
    ----------
    bool
        ``True`` only if this writer is recognized.

    See Also
    ----------
    :func:`is_writer_betse`, :func:`is_writer_mpl`
        Further details.
    '''

    return is_writer_betse(writer_name) and is_writer_mpl(writer_name)


@type_check
def is_writer_betse(writer_name: str) -> bool:
    '''
    ``True`` only if a matplotlib animation writer class (e.g.,
    :class:`FFMpegWriter`) with the passed name (e.g., ``ffmpeg``) is
    recognized by this application.

    Specifically, this function returns ``True`` only if this name is a key of
    the global :data:`WRITER_NAME_TO_COMMAND_BASENAME` dictionary of this
    submodule. Note this does *not* imply the external command required by this
    class (e.g., ``fmpeg``) to be installed on the current system.

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to test.

    Returns
    ----------
    bool
        ``True`` only if this writer is recognized by this application.
    '''

    return writer_name in WRITER_NAME_TO_COMMAND_BASENAME


@type_check
def is_writer_mpl(writer_name: str) -> bool:
    '''
    ``True`` only if a matplotlib animation writer class (e.g.,
    :class:`FFMpegWriter`) with the passed name (e.g., ``ffmpeg``) is
    registered with matplotlib, implying the external command required by this
    class (e.g., ``fmpeg``) to be installed on the current system.

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to test.

    Returns
    ----------
    bool
        ``True`` only if this writer is recognized by matplotlib.
    '''

    return writers.is_available(writer_name)

# ....................{ TESTERS ~ command                 }....................
@type_check
def is_writer_command(writer_basename: str) -> bool:
    '''
    ``True`` only if at least one matplotlib animation writer class (e.g.,
    :class:`MencoderWriter`) running the external command with the passed
    basename (e.g., ``mencoder``) is recognized by both this application *and*
    :mod:`matplotlib`, implying this command to be installed on the current
    system.

    Parameters
    ----------
    writer_basename : str
        Basename of the external command of the writer to test.

    Returns
    ----------
    bool
        ``True`` only if this command is recognized by both this application
        *and* :mod:`matplotlib`.
    '''

    # If this command is run by at least one writer...
    if writer_basename in WRITER_NAME_TO_COMMAND_BASENAME.reverse:
        # Tuple of the names of all writers running this command.
        writer_names = WRITER_NAME_TO_COMMAND_BASENAME.reverse[writer_basename]

        # If at least one such writer is recognized, return True.
        for writer_name in writer_names:
            if is_writer(writer_name):
                return True

    # Else, no such writers exist. Return False.
    return False


@type_check
def is_writer_command_codec(
    writer_basename: str, codec_name: StrOrNoneTypes) -> bool:
    '''
    ``True`` only if the matplotlib animation writer class running the external
    command with the passed basename (e.g., ``ffmpeg``) supports the video
    codec with the passed encoder-specific name (e.g., ``libx264``).

    Specifically, this function returns ``True`` only if this basename is:

    * ``ffmpeg`` and the ``ffmpeg -help encoder={codec_name}`` command
      succeeds.
    * ``avconv`` and the ``avconv -help encoder={codec_name}`` command
      succeeds.
    * ``mencoder``, the ``mencoder -ovc help`` command lists the
      Mencoder-specific ``lavc`` video codec and either:

      * ``ffmpeg`` is in the current ``${PATH}`` and recursively calling this
        function as ``is_writer_codec('ffmpeg', codec_name)`` returns ``True``.
      * ``ffmpeg`` is *not* in the current ``${PATH}``, in which case this
        function assumes the passed codec to be supported and returns ``True``.

    * Any other passed basename (e.g., ``convert``, implying ImageMagick) *and*
      the passed codec is ``None``. These basenames are assumed to *not*
      actually be video encoders and thus support no video codecs.

    Parameters
    ----------
    writer_basename : str
        Basename of the external command of the video encoder to test.
    codec_name : str
        Encoder-specific name of the codec to be tested for.

    Returns
    ----------
    bool
        ``True`` only if this encoder supports this codec.

    Raises
    ----------
    BetseMatplotlibException
        If this basename is either:

        * Unrecognized by this application.
        * Unregistered with :mod:`matplotlib`, implying
        * Not found as an external command in the current ``${PATH}``.
        * Mencoder and the ``mencoder -ovc help`` command fails to list the
          Mencoder-specific ``lavc`` video codec required by Matplotlib.
    '''

    # Log this detection attempt.
    logs.log_debug(
        'Detecting encoder "%s" codec "%s"...',
        writer_basename, codec_name)

    # Absolute path of this command.
    writer_filename = get_writer_command_filename(writer_basename)

    # For FFmpeg, detect this codec by capturing help documentation output by
    # the "ffmpeg" command for this codec and grepping this output for a string
    # stating this codec to be unrecognized. Sadly, this command reports
    # success rather than failure when this codec is unrecognized. (wut,
    # FFmpeg?)
    if writer_basename == 'ffmpeg':
        # List of all shell words comprising the command to be tested.
        ffmpeg_command_words = (
            writer_filename,
            '-help',
            'encoder=' + shellstr.shell_quote(codec_name),
        )

        # Help documentation for this codec captured from "ffmpeg".
        ffmpeg_codec_help = cmdrun.get_stdout_or_die(ffmpeg_command_words)

        # Return whether this documentation is suffixed by a string implying
        # this codec to be unrecognized or not. If this codec is unrecognized,
        # this documentation ends with the following line:
        #
        #     Codec '${codec_name}' is not recognized by FFmpeg.
        return not ffmpeg_codec_help.endswith("' is not recognized by FFmpeg.")
    # For Libav, detect this codec in the same exact manner as for FFmpeg.
    elif writer_basename == 'avconv':
        # List of all shell words comprising the command to be tested.
        avconv_command_words = (
            writer_filename,
            '-help',
            'encoder=' + shellstr.shell_quote(codec_name),
        )

        # Help documentation for this codec captured from "avconv".
        avconv_codec_help = cmdrun.get_stdout_or_die(avconv_command_words)
        # print('avconv_command_words: {}'.format(avconv_command_words))
        # print('avconv_codec_help: {}'.format(avconv_codec_help))

        # Return whether this documentation is suffixed by an indicative
        # string.
        return not avconv_codec_help.endswith("' is not recognized by Libav.")
    # For Mencoder, detect this codec by capturing help documentation output by
    # the "mencoder" command for *ALL* video codecs, grepping this output for
    # the "lavc" video codec required by matplotlib, and, if found, repeating
    # the above FFmpeg-specific logic to specifically detect this codec.
    elif writer_basename == 'mencoder':
        # Help documentation for all codecs captured from "mencoder".
        mencoder_codecs_help = cmdrun.get_stdout_or_die((
            writer_filename, '-ovc', 'help'))
        # print('mencoder codecs help: ' + mencoder_codecs_help)

        # If this output contains a line resembling the following, this
        # installation of Mencoder supports the requisite "lavc" codec:
        #     lavc     - libavcodec codecs - best quality!
        if regexes.is_match_line(
            text=mencoder_codecs_help, regex=r'^\s+lavc\s+'):
            # If the "ffmpeg" command is installed on the current system, query
            # that command for whether or not the passed codec is supported.
            # Note that the recursion bottoms out with this call, as the above
            # logic handling the FFmpeg writer does *NOT* recall this function.
            if is_writer_command('ffmpeg'):
                return is_writer_command_codec('ffmpeg', codec_name)
            # Else, "ffmpeg" is *NOT* in the ${PATH}. Since Mencoder implements
            # "lavc" codec support by linking against the "libavcodec" shared
            # library rather than calling the "ffmpeg" command, it's
            # technically permissible (albeit uncommon) for the "mencoder" but
            # not "ffmpeg" command to be in the ${PATH}. Hence, this does *NOT*
            # indicate a fatal error. This does, however, prevent us from
            # querying whether or not the passed codec is supported. In lieu of
            # sensible alternatives...
            else:
                # Log a non-fatal warning.
                logs.log_warning(
                    'Mencoder "libavcodec"-based video codec "{}" '
                    'possibly unavailable. Consider installing FFmpeg to '
                    'resolve this warning.'.format(codec_name))

                # Assume the passed codec to be supported.
                return True
        # Else, Mencoder fails to support the "lavc" codec. Raise an exception.
        else:
            raise BetseMatplotlibException(
                'Mencoder video codec "lavc" unavailable.')

    # For any other writer (e.g., ImageMagick), assume this writer to *NOT* be
    # a video encoder and hence support *NO* video codecs. In this case, return
    # True only if the passed codec is "None" -- signifying "no video codec."
    return codec_name is None

# ....................{ GETTERS                           }....................
@type_check
def get_writer_class(writer_name: str) -> ClassType:
    '''
    Matplotlib animation writer class (e.g., :class:`ImageMagickWriter`)
    registered with the passed name (e.g., ``imagemagick``).

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to obtain.

    Returns
    ----------
    ClassType
        Matplotlib animation writer class registered with this name.

    Raises
    ----------
    BetseMatplotlibException
        If this writer is unrecognized by either this application *or*
        :mod:`matplotlib`.
    '''

    # If this writer is unrecognized, raise an exception.
    die_unless_writer(writer_name)

    # Return this writer's class.
    return writers[writer_name]


@type_check
def get_writer_command_filename(writer_basename: str) -> str:
    '''
    Absolute path (e.g., ``/usr/bin/convert``) of the external command with the
    passed basename (e.g., ``convert``) run by a matplotlib animation writer
    class (e.g., :class:`ImageMagickWriter`).

    Parameters
    ----------
    writer_name : str
        Alphanumeric lowercase name of the writer to query.

    Returns
    ----------
    str
        Absolute path of this writer's command.

    Raises
    ----------
    BetseMatplotlibException
        If this writer is unrecognized by either this application *or*
        :mod:`matplotlib`.
    '''

    # If this command is unrecognized, raise an exception.
    die_unless_writer_command(writer_basename)

    # Tuple of the names of all writers running this command.
    writer_names = WRITER_NAME_TO_COMMAND_BASENAME.reverse[writer_basename]

    # Name of the first writer running this command. Since the absolute path of
    # this command is the same across all writers running this command, the
    # first writer is arbitrarily selected merely for simplicity.
    writer_name = writer_names[0]

    # This writer's class.
    writer_class = get_writer_class(writer_name)

    # Return the absolute path of this writer's command.
    return writer_class.bin_path()

# ....................{ GETTERS ~ first                   }....................
@type_check
def get_first_writer_name(writer_names: SequenceTypes) -> str:
    '''
    First name (e.g., ``imagemagick``) of the matplotlib animation writer class
    (e.g., :class:`ImageMagickWriter`) in the passed list recognized by both
    this application and :mod:`matplotlib` if any *or* raise an exception
    otherwise.

    This function iteratively searches for external commands in the same order
    as the passed list lists names.

    Parameters
    ----------
    writer_names : SequenceTypes
        List of the alphanumeric lowercase names of all writers to search for.

    Returns
    ----------
    str
        Alphanumeric lowercase name of the first such writer.

    Raises
    ----------
    BetseMatplotlibException
        If either:

        * Any writer in the passed list is unrecognized by this application.
        * No such writer is registered with :mod:`matplotlib`.
    '''

    # For the name of each passed writer...
    for writer_name in writer_names:
        # If this writer is unrecognized by BETSE, raise an exception. This
        # prevents BETSE from silently ignoring newly added writers not yet
        # recognized by BETSE.
        die_unless_writer_betse(writer_name)

        # If this writer is recognized by matplotlib, cease searching.
        if is_writer_mpl(writer_name):
            return writer_name

    # Else, no such command is in the ${PATH}. Raise an exception.
    raise BetseMatplotlibException(
        'Matplotlib animation video writers {} not found.'.format(
            strjoin.join_as_conjunction_double_quoted(*writer_names)))


@type_check
def get_first_codec_name(
    writer_name: str,
    container_filetype: str,
    codec_names: SequenceTypes,
) -> (str, NoneType):
    '''
    Name of the first video codec (e.g., ``libx264``) in the passed list
    supported by both the encoder with the passed matplotlib-specific name
    (e.g., ``ffmpeg``) and the container format with the passed filetype (e.g.,
    ``mkv``, ``mp4``) if that writer supports codecs (as most do), ``None`` if
    this writer supports no codecs (e.g., ``imagemagick``) and the passed list
    contains ``None``, *or* raise an exception otherwise (i.e., if no passed
    codecs are supported by both this writer and container format).

    Algorithm
    ----------
    This function iteratively searches for video codecs in the same order as
    listed in the passed list as follows:

    * If there are no remaining video codecs in this list to be examined, an
      exception is raised.
    * If the current video codec has the application-specific name ``auto``,
      the name of an intelligently selected codec supported by both this
      encoder and container if any is returned *or* an exception is raised
      otherwise (i.e., if no codecs are supported by both this encoder and
      container). Note that this codec's name rather than the
      application-specific name ``auto`` is returned. See this function's body
      for further commentary.
    * Else if the current video codec is supported by both this encoder and
      container, this codec's name is returned.
    * Else the next video codec in this list is examined.

    Parameters
    ----------
    writer_name : str
        Matplotlib-specific alphanumeric lowercase name of the video encoder to
        search for the passed codecs.
    container_filetype: str
        Filetype of the video container format to constrain this search to.
    codec_names: SequenceTypes
        Sequence of the encoder-specific names of all codecs to search for (in
        descending order of preference).

    Returns
    ----------
    str
        Name of the first codec in the passed list supported by both this
        encoder and container.

    Raises
    ----------
    BetseMatplotlibException
        If any of the following errors arise:

        * This writer is either:

          * Unrecognized by this application or :mod:`matplotlib`.
          * Not found as an external command in the current ``${PATH}``.

        * This container format is unsupported by this writer.
        * No codec whose name is in the passed list is supported by both this
          writer and this container format.

    See Also
    ----------
    :func:`is_writer`
        Tester validating this writer.
    '''

    # If this writer is unrecognized, raise an exception.
    die_unless_writer(writer_name)

    # Basename of this writer's command.
    writer_basename = WRITER_NAME_TO_COMMAND_BASENAME[writer_name]

    # Dictionary mapping from the filetype of each video container format to a
    # list of the names of all video codecs supported by this writer.
    container_filetype_to_codec_names = (
        WRITER_BASENAME_TO_CONTAINER_FILETYPE_TO_CODEC_NAMES[writer_basename])

    # If the passed container is unsupported by this writer, raise an exception.
    if container_filetype not in container_filetype_to_codec_names:
        raise BetseMatplotlibException(
            'Video container format "{}" unsupported by '
            'matplotlib animation video writer "{}".'.format(
                container_filetype, writer_name))

    # List of the names of all candidate codecs supported by both this encoder
    # and this container.
    codec_names_supported = container_filetype_to_codec_names[
        container_filetype]

    # List of the names of all candidate codecs to be detected below, with each
    # instance of "auto" in the original list of these names replaced by the
    # list of all codecs supported by both this encoder and container.
    codec_names_candidate = []

    # For the name of each candidate codec...
    for codec_name in codec_names:
        # If this is the BETSE-specific name "auto", append the names of all
        # codecs supported by both this encoder and container.
        if codec_name == 'auto':
            codec_names_candidate.extend(codec_names_supported)
        # Else, append only the name of this codec.
        else:
            codec_names_candidate.append(codec_name)

    # Log this detection attempt.
    logs.log_debug(
        'Detecting encoder "%s" codec from candidates: %r',
        writer_name, codec_names_candidate)

    # For the name of each preferred codec...
    for codec_name in codec_names_candidate:
        # If this encoder supports this codec *AND*...
        if is_writer_command_codec(writer_basename, codec_name):
            # Log this detection result.
            logs.log_debug(
                'Detected encoder "%s" codec "%s".', writer_name, codec_name)

            # If this container is not known to support this codec, log a
            # non-fatal warning. Since what this application thinks it knows
            # and what reality actually is need not coincide, this container
            # could actually still support this codec. Hence, this edge case
            # does *NOT* constitute a hard, fatal error.
            if codec_name not in codec_names_supported:
                logs.log_warning(
                    'Encoder "%s" container "%s" '
                    'not known to support codec "%s".',
                    writer_name, container_filetype, codec_name)

            # Return the name of this codec.
            return codec_name

    # Else, no passed codecs are supported by this combination of writer and
    # container format. Raise an exception.
    raise BetseMatplotlibException(
        'Codec(s) {} unsupported by '
        'encoder "{}" and/or container "{}".'.format(
            strjoin.join_as_conjunction_double_quoted(*codec_names),
            writer_name,
            container_filetype))

# ....................{ MAKERS                            }....................
#FIXME: Submit these fixes upstream, please. We'd rather not maintain them in
#perpuity, yet again.
@type_check
def make_writer(*args, cls: ClassType, **kwargs) -> MovieWriter:
    '''
    Create and return an instance of the passed matplotlib animation writer
    class, passing all remaining positional and keyword arguments to that
    class' ``__init__()`` method.

    This method wraps the ``__init__()`` method of the passed class with
    additional intelligence, effectively monkey-patching that method without
    unsafely doing so. Specifically, if:

    * A ``codec`` keyword argument is passed with value:

      * ``libx264``, the ``yuv420p`` pixel format is enabled. Note that the
        :meth:`FFMpegBase.output_args` method already does so onl* for the
        ``h264`` codec; since it should do so for both the ``libx264`` *and*
        ``h264`` codecs, this method coerces this behaviour.
      * ``libvpx``, the ``-auto-alt-ref 0`` command-line argument is passed.
        FFMpeg fails with the following fatal exception on failing to pass this
        parameter when selecting this popular codec:

        .. code:: bash

           MovieWriter.run: running command: ffmpeg -f rawvideo -vcodec rawvideo -s 1200x900 -pix_fmt rgba -r 15 -i pipe: -vcodec libvpx -b 1500k -metadata subject=Bioinformatics -metadata comment=Produced by BETSE. -metadata artist=BETSE -metadata genre=Bioinformatics -y /home/leycec/tmp/betse/sample_sim/RESULTS/init_1/anim_while_solving/Vmem/Vmem.mkv
           ffmpeg version 3.2.4 Copyright (c) 2000-2017 the FFmpeg developers
             built with gcc 4.9.4 (Gentoo 4.9.4 p1.0, pie-0.6.4)
           [betse] This run should take approximately 10.930000s to compute...
           [betse] Exporting "Vmem" frame 0 / 8...
           MovieWriter.grab_frame: Grabbing frame.
           Input #0, rawvideo, from 'pipe:':
             Duration: N/A, start: 0.000000, bitrate: 518400 kb/s
               Stream #0:0: Video: rawvideo (RGBA / 0x41424752), rgba, 1200x900, 518400 kb/s, 15 tbr, 15 tbn, 15 tbc
           Please use -b:a or -b:v, -b is ambiguous
           [libvpx @ 0x12bd960] v1.5.0
           [libvpx @ 0x12bd960] Transparency encoding with auto_alt_ref does not work
           Stream mapping:
             Stream #0:0 -> #0:0 (rawvideo (native) -> vp8 (libvpx))
           Error while opening encoder for output stream #0:0 - maybe incorrect parameters such as bit_rate, rate, width or height
           [betse] Exporting "Vmem" frame 1 / 8...
           MovieWriter.grab_frame: Grabbing frame.
           [betse] Exiting prematurely due to fatal error:

           BrokenPipeError: [Errno 32] Broken pipe

           Traceback (most recent call last):
             File "/home/leycec/py/betse/betse/cli/cliabc.py", line 163, in run
               profile_filename=self._profile_filename,
             File "<string>", line 65, in func_type_checked
             File "/home/leycec/py/betse/betse/util/py/profilers.py", line 130, in profile_callable
               profile_filename=profile_filename,
             File "/home/leycec/py/betse/betse/util/py/profilers.py", line 146, in _profile_callable_none
               return call(*args, **kwargs)
             File "/home/leycec/py/betse/betse/cli/climain.py", line 157, in _do
               return subcommand_method()
             File "/home/leycec/py/betse/betse/cli/climain.py", line 230, in _do_init
               return self._sim_runner.init()
             File "/home/leycec/py/betse/betse/science/simrunner.py", line 206, in init
               sim.run_sim_core(cells, p)
             File "/home/leycec/py/betse/betse/science/sim.py", line 929, in run_sim_core
               cells, p, time_steps, time_steps_sampled, anim_cells)
             File "<string>", line 109, in func_type_checked
             File "/home/leycec/py/betse/betse/science/sim.py", line 1238, in _run_sim_core_loop
               anim_cells.plot_frame(time_step=-1)
             File "<string>", line 15, in func_type_checked
             File "/home/leycec/py/betse/betse/science/visual/visabc.py", line 835, in plot_frame
               self._save_frame(time_step_absolute)
             File "/home/leycec/py/betse/betse/science/visual/visabc.py", line 1008, in _save_frame
               **self._writer_savefig_kwargs)
             File "/usr/lib64/python3.4/site-packages/matplotlib/animation.py", line 230, in grab_frame
               dpi=self.dpi, **savefig_kwargs)
             File "/usr/lib64/python3.4/site-packages/matplotlib/figure.py", line 1565, in savefig
               self.canvas.print_figure(*args, **kwargs)
             File "/usr/lib64/python3.4/site-packages/matplotlib/backend_bases.py", line 2232, in print_figure
               **kwargs)
             File "/usr/lib64/python3.4/site-packages/matplotlib/backends/backend_agg.py", line 519, in print_raw
               fileobj.write(renderer._renderer.buffer_rgba()

        For similar reports of the same underlying issue, see also:

        * https://stackoverflow.com/a/41311048/2809027
        * https://git.videolan.org/?p=ffmpeg.git;a=commitdiff;h=44bcb636c1a7052059330c3edbd900246e15ae60

    Parameters
    ----------
    cls : ClassType
        Matplotlib animation writer class to be instantiated.

    All remaining positional and keyword arguments are passed as is to this
    class' ``__init__()`` method.

    Returns
    ----------
    MovieWriter
        Matplotlib animation writer of this type instiated with these
        arguments.
    '''

    # If this class is *NOT* that of a Matplotlib animation writer, fail.
    classes.die_unless_subclass(subclass=cls, superclass=MovieWriter)

    # Name of the codec to be written if any or None otherwise.
    codec_name = kwargs.get('codec', None)

    # Additional command-line arguments to pass to the external command run by
    # this writer (e.g., "ffmpeg" for the "FFMpegWriter" class).
    writer_command_args = []

    # If this is an FFMpeg-specific writer passed a codec name of...
    if issubclass(cls, FFMpegBase) and codec_name is not None:
        # "libx264", force the "yuv420p" pixel format. See the docstring.
        if codec_name == 'libx264':
            writer_command_args.extend(['-pix_fmt', 'yuv420p'])
        # "libvpx", disable the "auto_alt_ref" feature. See the docstring.
        elif codec_name == 'libvpx':
            writer_command_args.extend(['-auto-alt-ref', '0'])

    # If passing this command additional command-line arguments, do so.
    if writer_command_args:
        kwargs['extra_args'] = writer_command_args

    # Create and return an instance of this class passed these arguments.
    return cls(*args, **kwargs)

# ....................{ MAIN                              }....................
# Initialize all global variables declared by this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **decorator** (i.e., callables dynamically wrapping arbitrary other
callables) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.io.log import logs
from betse.util.type.types import (
    type_check, CallableTypes, DecoratorTypes)
from functools import wraps

# ....................{ DECORATORS                        }....................
# While type-checking these types would probably be advisable, this decorator
# is currently passed non-callables by "py.test". While resolving that issue
# would itself probably be advisable, we simply cannot be bothered. Hence,
# these types remain blithely unchecked.
def decorator_identity(func: CallableTypes) -> CallableTypes:
    '''
    Identity decorator returning the decorated callable unmodified.
    '''

    return func

# ....................{ DECORATORS                        }....................
@type_check
def deprecated(func: CallableTypes) -> CallableTypes:
    '''
    Decorate the passed **callable** (e.g., function, method) to log a
    non-fatal warning on the first and *only* first call of this callable.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.call import callables

    # True only if this callable has had a deprecation warning logged.
    func.__is_deprecation_logged = False

    # Callable returned by this decorator below.
    @wraps(func)
    def _deprecated_inner(*args, **kwargs) -> object:
        # If this callable has *NOT* had a deprecation warning logged, do so.
        if not func.__is_deprecation_logged:
            # Prevent this warning from being logged more than once.
            func.__is_deprecation_logged = True

            # Capitalized human-readable string describing this callable.
            func_name = callables.to_str_capitalized(func)

            # Log this warning.
            logs.log_warning('%s deprecated.', func_name)

        # Call this callable.
        return func(*args, **kwargs)

    # Return this callable decorated to log this deprecating warning.
    return _deprecated_inner

# ....................{ CHAINERS                          }....................
@type_check
def decorator_chain(*decorators: DecoratorTypes) -> CallableTypes:
    '''
    Decorator chaining together (i.e., functionally composing) all passed
    decorators (in the reverse order).

    This function dynamically synthesizes a closure iteratively decorating any
    passed callable with each passed decorator. To preserve the expected
    semantics, that callable is decorated with:

    * The first passed decorator last.
    * The last passed decorator first.

    Parameters
    ----------
    decorators : tuple[DecoratorTypes]
        Tuple of all decorators to be chained together into a new decorator.

    Returns
    ----------
    CallableTypes
        Output decorator chaining together these input decorators.

    See Also
    ----------
    https://stackoverflow.com/a/5409569/2809027
        StackOverflow answer from which this implementation is inspired.
    '''

    # Closure to be returned as the new decorator.
    #
    # Note that the entirety of this function is reducible to this one-liner:
    #
    #     return lambda x: reduce(lambda y, f: f(y), decs, x)
    #
    # For maistainability, readability, and sanity, this function is instead
    # implemented with a traditional nested closure. Efficiency is irrelevant
    # in this context; moreover, the above one-liner is unlikely to yield
    # demonstrable efficiency gains.
    @type_check
    def _decorator_chain(func: CallableTypes) -> object:
        '''
        Iteratively decorate the passed callable with all decorators passed to
        the parent :func:`decorator_chain` function (in the reverse order),
        returning the callable object returned by the last such decorator.

        Parameters
        ----------
        call : CallableTypes
            Callable to be decorated.

        Returns
        ----------
        object
            Callable object decorating the input callable with these
            decorators.  While this object is typically a callable (i.e.,
            instance of :class:`CallableTypes`), this is *not* necessarily the
            case. For example, the :class:`property` decorator converts the
            passed unbound method into a data descriptor; strictly speaking,
            descriptors do *not* define the special ``__call__`` method and
            hence are *not* explicitly callable.
        '''

        # For each passed decorator (in reverse order)...
        for decorator in reversed(decorators):
            # Callable decorating the prior such callable with this decorator.
            func = decorator(func)

        # Return this decorated callable.
        return func

    # Return this new decorator.
    return _decorator_chain



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all YAML-backed wrapper subclasses.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.exceptions import BetseYamlException
# from betse.util.io.log import logs
from betse.util.type.types import type_check, MappingOrSequenceTypes, NoneType

# ....................{ SUPERCLASSES                      }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: To avoid diamond inheritance issues in the "YamlListABC" subclass,
# neither this subclass nor subclasses of this subclass should define any
# methods already defined by the "MutableSequence" API. This includes but is
# *NOT* limited to the following "MutableSequence" methods: append(), clear(),
# count(), extend(), index(), insert(), mro(), pop(), register(), remove(), and
# reverse().
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
class YamlABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **YAML-backed wrapper** (i.e., high-level object
    wrapping a low-level mapping or sequence of YAML-backed configuration
    settings) subclasses.

    Attributes
    ----------
    _conf : MappingOrSequenceOrNoneTypes
        Low-level container of related configuration settings loaded from and
        saved back to a YAML-formatted configuration file if the :meth:`load`
        method has been called *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self) -> None:
        '''
        Initialize this YAML wrapper.
        '''

        self._conf = None

    # ..................{ PROPERTIES ~ read-only            }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def is_loaded(self) -> bool:
        '''
        ``True`` only if this configuration is in the **loaded state** (i.e.,
        the :meth:`load` method has been called more recently than the
        :meth:`unload` method has been called).

        If ``True``, the :meth:`conf` property is safely accessible by callers.
        '''

        return self._conf is not None


    @property
    def conf(self) -> MappingOrSequenceTypes:
        '''
        Low-level container of related configuration settings loaded from and
        saved back to a YAML-formatted configuration file if the :meth:`load`
        method has been called more recently than the :meth:`unload` method has
        been called *or* raise an exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Else, this property is a container. Return this container.
        return self._conf

    # ..................{ EXCEPTIONS                        }..................
    def die_unless_loaded(self) -> None:
        '''
        Raise an exception unless this configuration is in the loaded state.

        Equivalently, this method raises an exception if this configuration is
        in the unloaded state.

        Raises
        ----------
        BetseYamlException
            If this configuration is in the unloaded state.

        See Also
        ----------
        :meth:`is_loaded`
            Further details.
        '''

        if not self.is_loaded:
            raise BetseYamlException(
                'YAML configuration not loaded '
                '(i.e., load() method not called yet).')

    # ..................{ LOADERS                           }..................
    @type_check
    def load(self, conf: MappingOrSequenceTypes) -> None:
        '''
        Associate this configuration with the passed YAML-backed container.

        Parameters
        ----------
        conf : MappingOrSequenceTypes
            Low-level container of related configuration settings loaded from
            and saved back to a YAML-formatted configuration file.
        '''

        # Classify this container.
        self._conf = conf


    def unload(self) -> None:
        '''
        Deassociate this configuration from its YAML-backed container if any
        *or* reduce to a noop otherwise.

        This method serves as a low-level safety mechanism ensuring that the
        next access of the :meth:`conf` property will raise a human-readable
        exception.
        '''

        self._conf = None

# ....................{ TYPES                             }....................
# Intended for use in callable type validation.
YamlABCOrNoneTypes = (YamlABC, NoneType)
'''
Tuple of both the YAML-backed configuration type *and* the type of the
singleton ``None`` object.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Blast-proof utilities.
'''

#FIXME: Consider offshoring this entire subpackage into a new independent
#Python project codenamed "Blast-proof utilities." PiPI project and package
#names deriving from this codename include:
#
#* "blapu". No existing PyPI packages. No prominent Google hits. Sorta ideal? My
#  own quandry is... well, let's face it: "Blapu." Really? That's awful.
#* "blaut". No existing PyPI packages. Many prominent Google hits. Less ideal?
#* "blauti". No existing PyPI packages. No prominent Google hits. *MOST IDEAL?*
#* "blapru". No existing PyPI packages. No prominent Google hits. Less ideal?
#* "blastu". No existing PyPI packages. Extremely prominent Google hits
#  pertaining to a technology festival hosted in Bravil, however. Non-ideal?
#
#I'm partial to "blauti," as of this writing. It just works. Only 5,000 Google
#hits effectively guarantees uniqueness. It also reads and sounds apropos.
#FIXME: Oh, yeah. We got this:
#
#* "brutil". Speaks for itself in waves of languid ludicrousness, doesn't it?



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Matplotlib-specific classes writing animations as frames.
'''

#FIXME: Consider contributing most or all of this submodule back to matplotlib.

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseMatplotlibException
from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from matplotlib.animation import writers, MovieWriter

# ....................{ CLASSES                            }....................
@writers.register('noop')
class NoopMovieWriter(MovieWriter):
    '''
    Matplotlib animation writer ignoring rather than writing animation frames.

    For drop-in use as an animation writer (e.g., to the :meth:`Animation.save`
    method), this class masquerades as a :class:`MovieWriter` subclass by the
    unique name of ``noop``. Nonetheless, no movies or files are written. This
    class is principally intended for use by callers:

    * Manually handling writing (e.g., by manually instantiating and calling
      methods of :class:`MovieWriter` subclasses).
    * Calling the :meth:`Animation.save` method, which obstructs manual
      handling of writing by:

      * Mandating that a writer be passed.
      * Automatically writing with that writer.

    Such callers may continue to manually handle writing by passing instances
    of this subclass to the :meth:`Animation.save` method.
    '''

    # ..................{ SUPERCLASS ~ class                 }..................
    @classmethod
    def isAvailable(cls):
        '''
        Notify users that this class does *not* depend on external commands and
        hence is *always* available.
        '''

        return True

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this writer.

        Specifically, the passed ``extra_args`` parameter if any is coerced to
        the empty list. Since this subclass forks no external process to write
        animation frames, all arguments to be passed to that process are
        ignorable.
        '''

        # Ignoring "extra_args" is essential. Failing to do so results in
        # exceptions in the superclass constructor, which when "extra_args" is
        # None attempts to access the non-existent "args_key" class attribute.
        kwargs['extra_args'] = list()
        super().__init__(*args, **kwargs)

    # ..................{ SUPERCLASS                         }..................
    def grab_frame(self, **kwargs) -> None:
        '''
        Prevent the superclass :meth:`grab_frame` method from writing this frame
        to the current sink for this writer.

        This writer fails to redefine the :meth:`_frame_sink` method and hence
        defaults to the sink provided by the superclass: stdin. Since stdin is
        typically an unsafe sink, this method circumvents the need to define a
        safe sink by reducing this method to a noop.
        '''

        pass


    def finish(self):
        '''
        Prevent the superclass :meth:`finish` method from attempting to capture
        output from an external process no longer forked by this subclass.
        '''

        pass


    def _run(self) -> None:
        '''
        Prevent the superclass :meth:`_run` method from attempting to fork an
        external process running a non-existent video encoding command.
        '''

        pass


# This writer shares sufficiently many similarities with the "noop" writer to
# warrant inheriting that writer.
@writers.register('image')
class ImageMovieWriter(NoopMovieWriter):
    '''
    Matplotlib animation writer writing animation frames to still image files
    rather than animated videos.

    For drop-in use as an animation writer (e.g., to the :meth:`Animation.save`
    method), this class masquerades as a :class:`MovieWriter` subclass by the
    unique name of ``image``. Nonetheless, no movie is written; only frames are
    written.

    Attributes
    -----------
    _frame_number : int
        0-based index of the next frame to be written.
    '''

    # ..................{ SUPERCLASS                        }..................
    def setup(self, *args, **kwargs) -> None:
        '''
        Prepare to write animation frames.

        This method is implicitly called by the superclass :meth:`saving`
        method implicitly called by the :meth:`Anim.save` method. Note that,
        unfortunately, the design of both methods prohibits this method from
        accepting subclass-specific parameters.

        Parameters
        -----------
        outfile : str
            :meth:`str.format`-formatted template describing the filenames of
            the resulting frame images. This template is subject to the
            following :meth:`str.format` formatting:

            * The first ``{``- and ``}``-delimited substring (e.g., ``{:07d}``)
              will be replaced by the 0-based index of the current frame. If
              this substring does *not* exist, an exception is raised.

        See the superclass :meth:`setup` method for details on all other
        parameters.

        See Also
        -----------
        https://www.python.org/dev/peps/pep-3101
            For details on format specifiers.
        '''

        super().setup(*args, **kwargs)

        # Frame number of the next frame to be written.
        self._frame_number = 0

        # If this filename template is malformed, raise an exception.
        if not ('{' in self.outfile and '}' in self.outfile):
            raise BetseMatplotlibException(
                'Frame filename template "{}" contains no "{{"- and "}}"-'
                'delimited format specifier.'.format(self.outfile))

        # Output filetype. Override the superclass' awkward choice of "rgba" as
        # output filetype default.
        self.frame_format = pathnames.get_filetype_undotted_or_none(
            self.outfile)

        # List of all output filetypes supported by this class.
        #
        # Since this class serializes frames by calling the savefig() function
        # *AND* since that function supports all image filetypes supported by
        # the current backend, this syllogistically follows. (Q.E.D.)
        out_filetypes_supported = self.fig.canvas.get_supported_filetypes()

        # If this filetype is unsupported, raise an exception.
        if self.frame_format not in out_filetypes_supported:
            raise BetseMatplotlibException(
                'Frame filetype "{}" unsupported by the '
                'current Matplotlib backend (i.e., not in "{}").'.format(
                    self.frame_format, str(out_filetypes_supported)))

        # Parent directory of all output files.
        out_dirname = pathnames.get_dirname(self.outfile)

        # Create this directory if needed.
        dirs.make_parent_unless_dir(out_dirname)


    def grab_frame(self, **kwargs) -> None:
        '''
        Write the next frame for the current figure to the image file defined
        by the current filename template.

        The high-level :meth:`Animation.save` method unpacks and passes all
        key-value pairs of the optional ``savefig_kwargs`` dictionary argument
        to this method as keyword arguments. This method then passes all passed
        keyword arguments to the :meth:`Figure.savefig` method.
        '''

        # Leverage similar code as our superclass with exception of:
        #
        # * *NOT* catching and logging exceptions. The superclass
        #   implementation catches and logs exceptions by deferring to the
        #   output of an external process no longer forked by this subclass.
        # * Passing a filename rather than open file handle to the
        #   self.fig.savefig() method. While we could reimplement the
        #   superclass _frame_sink() method to do so, this seems simpler and
        #   more reliable.

        # Filename of the current frame to be written.
        frame_filename = self.outfile.format(self._frame_number)

        # Log this attempt. Actually, forget it. Far too verbose for now.
        # logs.log_debug(
        #     'ImageMovieWriter.grab_frame: saving frame %d to fname=%s',
        #     self._frame_number, frame_filename)
        # logs.log_debug(
        #     'Saving frame %d with options: %s', self._frame_number, str(kwargs))

        # Increment the number of the next frame to be written *AFTER* logging.
        self._frame_number += 1

        # Write the current frame.
        self.fig.savefig(
            # The public matplotlib API expects the first argument to this
            # method to be passed positionally rather than as a keyword
            # argument. We know this both because:
            #
            # * This argument is *ALWAYS* passed positionally by the matplotlib
            #   codebase itself.
            # * The name of this argument has changed between matplotlib
            #   versions (notably, from "filename" to "fname"), preventing this
            #   argument from being reliably passed as a keyword argument. To
            #   preserve forward compatibility with multiple matplotlib
            #   versions, this argument *MUST* be passed positionally.
            frame_filename,

            # All remaining arguments are expected to be keyword arguments.
            format=self.frame_format,
            dpi=self.dpi,
            **kwargs
        )
